"""Serializable selector-related classes."""

from __future__ import annotations

import logging
from typing import Any, Optional, Tuple, TypeVar

from trulens.core._utils.pycompat import TypeAlias
from trulens.core.utils import deprecation as deprecation_utils
from trulens.core.utils import serial as serial_utils

T = TypeVar("T")

logger = logging.getLogger(__name__)


class Select:
    """
    Utilities for creating selectors using Lens and aliases/shortcuts.
    """

    Lens: TypeAlias = serial_utils.Lens

    Query: TypeAlias = Lens  # Deprecate this alias in the future.

    Tru: Lens = Lens()
    """Selector for the tru wrapper (TruLlama, TruChain, etc.)."""

    Record: Lens = Lens().__record__
    """Selector for the record."""

    App: Lens = Lens().__app__
    """Selector for the app."""

    RecordInput: Lens = Record.main_input
    """Selector for the main app input."""

    RecordOutput: Lens = Record.main_output
    """Selector for the main app output."""

    RecordCalls: Lens = Record.app  # type: ignore
    """Selector for the calls made by the wrapped app.

    Laid out by path into components.
    """

    RecordCall: Lens = Record.calls[-1]
    """Selector for the first called method (last to return)."""

    RecordArgs: Lens = RecordCall.args
    """Selector for the whole set of inputs/arguments to the first called / last method call."""

    RecordRets: Lens = RecordCall.rets
    """Selector for the whole output of the first called / last returned method call."""

    _PREFIXES = [
        ("Select.RecordInput", RecordInput),
        ("Select.RecordOutput", RecordOutput),
        ("Select.RecordArgs", RecordArgs),
        ("Select.RecordRets", RecordRets),
        ("Select.RecordCalls", RecordCalls),
        ("Select.RecordCall", RecordCall),
        ("Select.Record", Record),
        ("Select.App", App),
    ]
    """All prefixes/shorthands defined in this class.

    Make sure this list is sorted by longest prefix first as some prefixes are
    prefixes of others.
    """

    @staticmethod
    def path_and_method(select: Select.Lens) -> Tuple[Select.Lens, str]:
        """
        If `select` names in method as the last attribute, extract the method name
        and the selector without the final method name.
        """

        if len(select.path) == 0:
            raise ValueError(
                "Given selector is empty so does not name a method."
            )

        firsts = select.path[:-1]
        last = select.path[-1]

        if not isinstance(last, serial_utils.StepItemOrAttribute):
            raise ValueError(
                "Last part of selector is not an attribute so does not name a method."
            )

        method_name = last.get_item_or_attribute()
        path = Select.Lens(path=firsts)

        return path, method_name

    @staticmethod
    def dequalify(lens: Select.Lens) -> Select.Lens:
        """If the given selector qualifies record or app, remove that qualification."""

        if len(lens.path) == 0:
            return lens

        if Select.Record.is_prefix_of(lens) or Select.App.is_prefix_of(lens):
            return Select.Lens(path=lens.path[len(Select.Record) :])

        return lens

    @deprecation_utils.staticmethod_renamed(
        "trulens.core.app.base.App.select_context"
    )
    @staticmethod
    def context(app: Optional[Any] = None) -> Lens:
        """DEPRECATED: Select the context (retrieval step outputs) of the given
        app."""

        if app is None:
            raise ValueError("App must be given to select context.")

        return app.select_context(app)

    @staticmethod
    def for_record(lens: Select.Lens) -> Lens:
        """Add the Record prefix to the beginning of the given lens."""

        return Select.Lens(path=Select.Record.path + lens.path)

    @staticmethod
    def for_app(lens: Select.Lens) -> Lens:
        """Add the App prefix to the beginning of the given lens."""

        return Select.Lens(path=Select.App.path + lens.path)

    @staticmethod
    def render_for_dashboard(lens: Select.Lens) -> str:
        """Render the given lens for use in dashboard to help user specify feedback functions."""

        if len(lens) == 0:
            return "Select.Lens()"

        ret = ""
        rest = None

        for prefix_name, prefix_lens in Select._PREFIXES:
            if prefix_lens.is_prefix_of(lens):
                ret = prefix_name
                rest = lens.path[len(prefix_lens) :]
                break

        if rest is None:
            rest = lens.path

        for step in rest:
            ret += repr(step)

        return f"{ret}"



from __future__ import annotations

from collections import defaultdict
from concurrent import futures
from datetime import datetime
import inspect
import json
import logging
from multiprocessing import Process
import threading
from threading import Thread
from time import sleep
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)
import warnings

import pandas
import pydantic
from trulens.core import experimental as core_experimental
from trulens.core._utils import optional as optional_utils
from trulens.core._utils.pycompat import Future  # code style exception
from trulens.core.database import connector as core_connector
from trulens.core.feedback import feedback as core_feedback
from trulens.core.metric import metric as core_metric
from trulens.core.otel.utils import is_otel_tracing_enabled
from trulens.core.schema import app as app_schema
from trulens.core.schema import dataset as dataset_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import groundtruth as groundtruth_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import deprecation as deprecation_utils
from trulens.core.utils import imports as import_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils
from trulens.core.utils import threading as threading_utils
from trulens.experimental.otel_tracing import _feature as otel_tracing_feature
from trulens.otel.semconv.trace import ResourceAttributes
from trulens.otel.semconv.trace import SpanAttributes

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import TracerProvider
    from opentelemetry.sdk.trace.export import SpanExporter
    from trulens.core import app as base_app
    from trulens.core.otel.recording import Record

tqdm = None
with import_utils.OptionalImports(messages=optional_utils.REQUIREMENT_TQDM):
    from tqdm.auto import tqdm

logger = logging.getLogger(__name__)


class TruSession(
    core_experimental._WithExperimentalSettings, python_utils.PydanticSingleton
):
    """TruSession is the main class that provides an entry points to trulens.

    TruSession lets you:

    - Log app prompts and outputs
    - Log app Metadata
    - Run and log feedback functions
    - Run streamlit dashboard to view experiment results

    By default, all data is logged to the current working directory to
    `"default.sqlite"`. Data can be logged to a SQLAlchemy-compatible url
    referred to by `database_url`.

    Supported App Types:
        [TruChain][trulens.apps.langchain.TruChain]: Langchain
            apps.

        [TruLlama][trulens.apps.llamaindex.TruLlama]: Llama Index
            apps.

        [TruRails][trulens.apps.nemo.TruRails]: NeMo Guardrails apps.

        [TruBasicApp][trulens.apps.basic.TruBasicApp]:
            Basic apps defined solely using a function from `str` to `str`.

        [TruApp][trulens.apps.app.TruApp]:
            Custom apps containing custom structures and methods. Requires
            annotation of methods to instrument.

        [TruVirtual][trulens.apps.virtual.TruVirtual]: Virtual
            apps that do not have a real app to instrument but have a virtual
            structure and can log existing captured data as if they were trulens
            records.

    Args:
        connector: Database Connector to use. If not provided, a default
            [DefaultDBConnector][trulens.core.database.connector.default.DefaultDBConnector]
            is created.

        experimental_feature_flags: Experimental feature flags.

        **kwargs: All other arguments are used to initialize
            [DefaultDBConnector][trulens.core.database.connector.default.DefaultDBConnector].
            Mutually exclusive with `connector`.
    """

    model_config = pydantic.ConfigDict(
        arbitrary_types_allowed=True, extra="forbid"
    )

    RETRY_RUNNING_SECONDS: float = 60.0
    """How long to wait (in seconds) before restarting a feedback function that
    has already started

    A feedback function execution that has started may have stalled or failed in
    a bad way that did not record the failure.

    See also:
        [start_evaluator][trulens.core.session.TruSession.start_evaluator]

        [DEFERRED][trulens.core.schema.feedback.FeedbackMode.DEFERRED]
    """

    RETRY_FAILED_SECONDS: float = 5 * 60.0
    """How long to wait (in seconds) to retry a failed feedback function run."""

    DEFERRED_NUM_RUNS: int = 32
    """Number of futures to wait for when evaluating deferred feedback functions."""

    RECORDS_BATCH_TIMEOUT_IN_SEC: int = 10
    """Time to wait before inserting a batch of records into the database."""

    GROUND_TRUTHS_BATCH_SIZE: int = 100
    """Time to wait before inserting a batch of ground truths into the database."""

    _evaluator_stop: Optional[threading.Event] = pydantic.PrivateAttr(None)
    """Event for stopping the deferred evaluator which runs in another thread."""

    _evaluator_proc: Optional[Union[Process, Thread]] = pydantic.PrivateAttr(
        None
    )

    _dashboard_urls: Optional[str] = pydantic.PrivateAttr(None)

    _dashboard_proc: Optional[Process] = pydantic.PrivateAttr(None)

    _tunnel_listener_stdout: Optional[Thread] = pydantic.PrivateAttr(None)

    _tunnel_listener_stderr: Optional[Thread] = pydantic.PrivateAttr(None)

    _dashboard_listener_stdout: Optional[Thread] = pydantic.PrivateAttr(None)

    _dashboard_listener_stderr: Optional[Thread] = pydantic.PrivateAttr(None)

    connector: Optional[core_connector.DBConnector] = pydantic.Field(
        None, exclude=True
    )
    """Database Connector to use. If not provided, a default is created and
    used."""

    _experimental_otel_exporter: Optional[SpanExporter] = pydantic.PrivateAttr(
        None
    )

    _experimental_tracer_provider: Optional[TracerProvider] = (
        pydantic.PrivateAttr(None)
    )

    @property
    def experimental_otel_exporter(
        self,
    ) -> Optional[SpanExporter]:
        """EXPERIMENTAL(otel_tracing): OpenTelemetry SpanExporter to send spans
        to.

        Only works if the trulens.core.experimental.Feature.OTEL_TRACING flag is
        set. The setter will set and lock the flag as enabled.
        """

        return self._experimental_otel_exporter

    def force_flush(self, timeout_millis: int = 300000) -> bool:
        """
        Force flush the OpenTelemetry exporters.

        Args:
            timeout_millis: The maximum amount of time to wait for spans to be
                processed.

        Returns:
            False if the timeout is exceeded, feature is not enabled, or the provider doesn't exist, True otherwise.
        """

        if (
            not self.experimental_feature(
                core_experimental.Feature.OTEL_TRACING
            )
            or self._experimental_tracer_provider is None
        ):
            return False

        return self._experimental_tracer_provider.force_flush(timeout_millis)

    def __str__(self) -> str:
        return f"TruSession({self.connector})"

    # For WithExperimentalSetttings mixin's WithIdentString requirement
    def _ident_str(self) -> str:
        if self.connector is None:
            return "TruSession(no connector)"

        return self.connector._ident_str()

    def __init__(
        self,
        connector: Optional[core_connector.DBConnector] = None,
        experimental_feature_flags: Optional[
            Union[
                Mapping[core_experimental.Feature, bool],
                List[core_experimental.Feature],
            ]
        ] = None,
        _experimental_otel_exporter: Optional[SpanExporter] = None,
        **kwargs,
    ):
        if python_utils.safe_hasattr(self, "connector"):
            # Already initialized by SingletonByName mechanism. Give warning if
            # any option was specified (not None) as it will be ignored.
            if connector is not None:
                logger.warning(
                    "TruSession was already initialized. Cannot change database configuration after initialization."
                )
            return
        connector_args = {
            k: v
            for k, v in kwargs.items()
            if k
            in inspect.signature(
                core_connector.DefaultDBConnector.__init__
            ).parameters
        }
        self_args = {k: v for k, v in kwargs.items() if k not in connector_args}

        if connector_args and connector is not None:
            extra_keys = ", ".join(
                map(lambda s: "`" + s + "`", connector_args.keys())
            )
            raise ValueError(
                f"Cannot provide both `connector` and connector argument(s) {extra_keys}."
            )

        connector = connector or core_connector.DefaultDBConnector(
            **connector_args
        )

        super().__init__(
            connector=connector,
            **self_args,
        )

        # for WithExperimentalSettings mixin
        self.experimental_set_features(experimental_feature_flags)

        if (
            _experimental_otel_exporter is not None
            and not self.experimental_feature(
                core_experimental.Feature.OTEL_TRACING
            )
        ):
            raise ValueError(
                "Cannot supply `_experimental_otel_exporter` without enabling OTEL tracing!"
            )
        if self.experimental_feature(core_experimental.Feature.OTEL_TRACING):
            otel_tracing_feature._FeatureSetup.assert_optionals_installed()

            from trulens.experimental.otel_tracing.core.session import (
                _TruSession,
            )

            _TruSession._set_up_otel_exporter(
                self, connector, _experimental_otel_exporter
            )

            _TruSession._track_costs()

    def App(self, *args, app: Optional[Any] = None, **kwargs) -> base_app.App:
        """Create an App from the given App constructor arguments by guessing
        which app type they refer to.

        This method intentionally prints out the type of app being created to
        let user know in case the guess is wrong.
        """
        if app is None:
            # If app is not given as a keyword argument, check the positional args.

            if len(args) == 0:
                # Basic app can be specified using the text_to_text key argument.
                if "text_to_text" in kwargs:
                    from trulens.apps import basic

                    return basic.TruBasicApp(
                        *args, connector=self.connector, **kwargs
                    )

                raise ValueError("No app provided.")

            # Otherwise the app must be the first positional arg.
            app, args = args[0], args[1:]

        # Check for optional app types.
        if app.__module__.startswith("langchain"):
            with import_utils.OptionalImports(
                messages=optional_utils.REQUIREMENT_APPS_LANGCHAIN
            ):
                from trulens.apps.langchain import tru_chain

            print(f"{text_utils.UNICODE_SQUID} Instrumenting LangChain app.")
            return tru_chain.TruChain(
                *args, app=app, connector=self.connector, **kwargs
            )

        elif app.__module__.startswith("langgraph"):
            with import_utils.OptionalImports(
                messages=optional_utils.REQUIREMENT_APPS_LANGGRAPH
            ):
                from trulens.apps.langgraph import tru_graph

            print(f"{text_utils.UNICODE_SQUID} Instrumenting LangGraph app.")
            return tru_graph.TruGraph(
                *args, app=app, connector=self.connector, **kwargs
            )

        elif app.__module__.startswith("llamaindex"):
            with import_utils.OptionalImports(
                messages=optional_utils.REQUIREMENT_APPS_LLAMA
            ):
                from trulens.apps.llamaindex import tru_llama

            print(f"{text_utils.UNICODE_SQUID} Instrumenting LlamaIndex app.")
            return tru_llama.TruLlama(
                *args, app=app, connector=self.connector, **kwargs
            )

        elif app.__module__.startswith("nemoguardrails"):
            with import_utils.OptionalImports(
                messages=optional_utils.REQUIREMENT_APPS_NEMO
            ):
                from trulens.apps.nemo import tru_rails

            print(
                f"{text_utils.UNICODE_SQUID} Instrumenting NeMo GuardRails app."
            )

            return tru_rails.TruRails(
                *args, app=app, connector=self.connector, **kwargs
            )

        # Check for virtual. Either VirtualApp or JSON app arg.
        from trulens.apps import virtual
        from trulens.core.utils import serial as serial_utils

        if isinstance(app, virtual.VirtualApp) or serial_utils.is_json(app):
            print(f"{text_utils.UNICODE_SQUID} Instrumenting virtual app.")
            return virtual.TruVirtual(
                *args, app=app, connector=self.connector, **kwargs
            )

        # Check for basic. Either TruWrapperApp or the text_to_text arg. Unsure
        # what we want to do if they provide both. Let's TruBasicApp handle it.
        from trulens.apps import basic

        if isinstance(app, basic.TruWrapperApp) or "text_to_text" in kwargs:
            print(f"{text_utils.UNICODE_SQUID} Instrumenting basic app.")

            return basic.TruBasicApp(
                *args, app=app, connector=self.connector, **kwargs
            )

        # If all else fails, assume it is a custom app.
        print(f"{text_utils.UNICODE_SQUID} Instrumenting custom app.")
        from trulens.apps import custom

        return custom.TruCustomApp(
            *args, app=app, connector=self.connector, **kwargs
        )

    @deprecation_utils.method_renamed("TruSession.App")
    def Basic(self, *args, **kwargs) -> base_app.App:
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.App][trulens.core.session.TruSession.App]
            instead.
        """
        from trulens.apps.basic import TruBasicApp

        return TruBasicApp(*args, connector=self.connector, **kwargs)

    @deprecation_utils.method_renamed("TruSession.App")
    def Custom(self, *args, **kwargs) -> base_app.App:
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.App][trulens.core.session.TruSession.App]
            instead.
        """
        from trulens.apps.custom import TruCustomApp

        return TruCustomApp(*args, connector=self.connector, **kwargs)

    @deprecation_utils.method_renamed("TruSession.App")
    def Virtual(self, *args, **kwargs) -> base_app.App:
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.App][trulens.core.session.TruSession.App]
            instead.
        """
        from trulens.apps.virtual import TruVirtual

        return TruVirtual(*args, connector=self.connector, **kwargs)

    @deprecation_utils.method_renamed("TruSession.App")
    def Chain(self, *args, **kwargs) -> base_app.App:
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.App][trulens.core.session.TruSession.App]
            instead.
        """
        with import_utils.OptionalImports(
            messages=optional_utils.REQUIREMENT_APPS_LANGCHAIN
        ):
            from trulens.apps.langchain import tru_chain

        return tru_chain.TruChain(*args, connector=self.connector, **kwargs)

    @deprecation_utils.method_renamed("TruSession.App")
    def Llama(self, *args, **kwargs) -> base_app.App:
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.App][trulens.core.session.TruSession.App]
            instead.
        """
        with import_utils.OptionalImports(
            messages=optional_utils.REQUIREMENT_APPS_LLAMA
        ):
            from trulens.apps.llamaindex import tru_llama

        return tru_llama.TruLlama(*args, connector=self.connector, **kwargs)

    @deprecation_utils.method_renamed("TruSession.App")
    def Rails(self, *args, **kwargs) -> base_app.App:
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.App][trulens.core.session.TruSession.App]
            instead.
        """
        with import_utils.OptionalImports(
            messages=optional_utils.REQUIREMENT_APPS_NEMO
        ):
            from trulens.apps.nemo import tru_rails

        return tru_rails.TruRails(*args, connector=self.connector, **kwargs)

    @deprecation_utils.method_renamed("trulens.dashboard.run.find_unused_port")
    def find_unused_port(self, *args, **kwargs):
        """
        !!! warning "Deprecated"
            Use
            [trulens.dashboard.run.find_unused_port][trulens.dashboard.run.find_unused_port]
            instead.
        """
        from trulens.dashboard.run import find_unused_port

        return find_unused_port(*args, **kwargs)

    @deprecation_utils.method_renamed("trulens.dashboard.run.run_dashboard")
    def run_dashboard(self, *args, **kwargs):
        """
        !!! warning "Deprecated"
            Use
            [trulens.dashboard.run.run_dashboard][trulens.dashboard.run.run_dashboard]
            instead.
        """
        from trulens.dashboard.run import run_dashboard

        return run_dashboard(*args, session=self, **kwargs)

    @deprecation_utils.method_renamed("trulens.dashboard.run.run_dashboard")
    def start_dashboard(self, *args, **kwargs):
        """
        !!! warning "Deprecated"
            Use
            [trulens.dashboard.run.run_dashboard][trulens.dashboard.run.run_dashboard]
            instead.
        """
        from trulens.dashboard.run import run_dashboard

        return run_dashboard(*args, session=self, **kwargs)

    @deprecation_utils.method_renamed("trulens.dashboard.run.stop_dashboard")
    def stop_dashboard(self, *args, **kwargs):
        """
        !!! warning "Deprecated"
            Use
            [trulens.dashboard.run.stop_dashboard][trulens.dashboard.run.stop_dashboard]
            instead.
        """
        from trulens.dashboard.run import stop_dashboard

        return stop_dashboard(*args, session=self, **kwargs)

    @deprecation_utils.method_renamed("TruSession.connector.db.insert_record")
    def update_record(self, *args, **kwargs):
        """
        !!! warning "Deprecated"
            Use
            [trulens.core.session.TruSession.connector][trulens.core.session.TruSession.connector] [.db.insert_record][trulens.core.database.base.DB.insert_record]
            instead.
        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        assert self.connector is not None
        return self.connector.db.insert_record(*args, **kwargs)

    def reset_database(self):
        """Reset the database. Clears all tables.

        See [DB.reset_database][trulens.core.database.base.DB.reset_database].
        """
        self.connector.reset_database()

    def migrate_database(self, **kwargs: Dict[str, Any]):
        """Migrates the database.

        This should be run whenever there are breaking changes in a database
        created with an older version of _trulens_.

        Args:
            **kwargs: Keyword arguments to pass to
                [migrate_database][trulens.core.database.base.DB.migrate_database]
                of the current database.

        See [DB.migrate_database][trulens.core.database.base.DB.migrate_database].
        """
        print(f"{text_utils.UNICODE_SQUID} Migrating DB ...")
        self.connector.migrate_database(**kwargs)

    def add_record(
        self, record: Optional[record_schema.Record] = None, **kwargs: dict
    ) -> types_schema.RecordID:
        """Add a record to the database.

        Args:
            record: The record to add.

            **kwargs: [Record][trulens.core.schema.record.Record] fields to add to the
                given record or a new record if no `record` provided.

        Returns:
            Unique record identifier [str][] .

        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        return self.connector.add_record(record=record, **kwargs)

    def add_record_nowait(
        self,
        record: record_schema.Record,
    ) -> None:
        """Add a record to the queue to be inserted in the next batch."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        return self.connector.add_record_nowait(record)

    def run_feedback_functions(
        self,
        record: record_schema.Record,
        feedback_functions: Sequence[core_metric.Metric],
        app: Optional[app_schema.AppDefinition] = None,
        wait: bool = True,
    ) -> Union[
        Iterable[feedback_schema.FeedbackResult],
        Iterable[Future[feedback_schema.FeedbackResult]],
    ]:
        """Run a collection of feedback functions and report their result.

        Args:
            record: The record on which to evaluate the feedback
                functions.

            app: The app that produced the given record.
                If not provided, it is looked up from the given database `db`.

            feedback_functions: A collection of feedback
                functions to evaluate.

            wait: If set (default), will wait for results
                before returning.

        Yields:
            One result for each element of `feedback_functions` of
                [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] if `wait`
                is enabled (default) or [Future][concurrent.futures.Future] of
                [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] if `wait`
                is disabled.
        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")

        if not isinstance(record, record_schema.Record):
            raise ValueError(
                "`record` must be a `trulens.core.schema.record.Record` instance."
            )

        if not isinstance(feedback_functions, Sequence):
            raise ValueError("`feedback_functions` must be a sequence.")

        if not all(
            isinstance(ffunc, core_feedback.Feedback)
            for ffunc in feedback_functions
        ):
            raise ValueError(
                "`feedback_functions` must be a sequence of `trulens.core.Feedback` instances."
            )

        if not (app is None or isinstance(app, app_schema.AppDefinition)):
            raise ValueError(
                "`app` must be a `trulens.core.schema.app.AppDefinition` instance."
            )

        if not isinstance(wait, bool):
            raise ValueError("`wait` must be a bool.")

        future_feedback_map: Dict[
            Future[feedback_schema.FeedbackResult], core_feedback.Feedback
        ] = {
            p[1]: p[0]
            for p in app_schema.AppDefinition._submit_feedback_functions(
                record=record,
                feedback_functions=feedback_functions,
                connector=self.connector,
                app=app,
            )
        }

        if wait:
            # In blocking mode, wait for futures to complete.
            for fut_result in futures.as_completed(future_feedback_map.keys()):
                # TODO: Do we want a version that gives the feedback for which
                # the result is being produced too? This is more useful in the
                # Future case as we cannot check associate a Future result to
                # its feedback before result is ready.

                # yield (future_feedback_map[fut_result], fut_result.result())
                yield fut_result.result()

        else:
            # In non-blocking, return the futures instead.
            for fut_result, _ in future_feedback_map.items():
                # TODO: see prior.

                # yield (feedback, fut_result)
                yield fut_result

    def add_app(self, app: app_schema.AppDefinition) -> types_schema.AppID:
        """
        Add an app to the database and return its unique id.

        Args:
            app: The app to add to the database.

        Returns:
            A unique app identifier [str][].

        """
        return self.connector.add_app(app=app)

    def delete_app(self, app_id: types_schema.AppID) -> None:
        """
        Deletes an app from the database based on its app_id.

        Args:
            app_id (schema.AppID): The unique identifier of the app to be deleted.
        """
        return self.connector.delete_app(app_id=app_id)

    def add_feedback(
        self,
        feedback_result_or_future: Optional[
            Union[
                feedback_schema.FeedbackResult,
                Future[feedback_schema.FeedbackResult],
            ]
        ] = None,
        **kwargs: dict,
    ) -> types_schema.FeedbackResultID:
        """Add a single feedback result or future to the database and return its unique id.

        Args:
            feedback_result_or_future: If a [Future][concurrent.futures.Future]
                is given, call will wait for the result before adding it to the
                database. If `kwargs` are given and a
                [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] is also
                given, the `kwargs` will be used to update the
                [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] otherwise a
                new one will be created with `kwargs` as arguments to its
                constructor.

            **kwargs: Fields to add to the given feedback result or to create a
                new [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] with.

        Returns:
            A unique result identifier [str][].

        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        return self.connector.add_feedback(
            feedback_result_or_future=feedback_result_or_future, **kwargs
        )

    def add_feedbacks(
        self,
        feedback_results: Iterable[
            Union[
                feedback_schema.FeedbackResult,
                Future[feedback_schema.FeedbackResult],
            ]
        ],
    ) -> List[types_schema.FeedbackResultID]:
        """Add multiple feedback results to the database and return their unique ids.

        Args:
            feedback_results: An iterable with each iteration being a [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] or
                [Future][concurrent.futures.Future] of the same. Each given future will be waited.

        Returns:
            List of unique result identifiers [str][] in the same order as input
                `feedback_results`.
        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        return self.connector.add_feedbacks(feedback_results=feedback_results)

    def get_app(
        self, app_id: types_schema.AppID
    ) -> Optional[serial_utils.JSONized[app_schema.AppDefinition]]:
        """Look up an app from the database.

        This method produces the JSON-ized version of the app. It can be deserialized back into an [AppDefinition][trulens.core.schema.app.AppDefinition] with [model_validate][pydantic.BaseModel.model_validate]:

        Example:
            ```python
            from trulens.core.schema import app
            app_json = session.get_app(app_id="app_hash_85ebbf172d02e733c8183ac035d0cbb2")
            app = app.AppDefinition.model_validate(app_json)
            ```

        Warning:
            Do not rely on deserializing into [App][trulens.core.app.App] as
            its implementations feature attributes not meant to be deserialized.

        Args:
            app_id: The unique identifier [str][] of the app to look up.

        Returns:
            JSON-ized version of the app.
        """

        return self.connector.get_app(app_id)

    def get_apps(self) -> List[serial_utils.JSONized[app_schema.AppDefinition]]:
        """Look up all apps from the database.

        Returns:
            A list of JSON-ized version of all apps in the database.

        Warning:
            Same Deserialization caveats as [get_app][trulens.core.session.TruSession.get_app].
        """

        return self.connector.get_apps()

    def get_records_and_feedback(
        self,
        app_ids: Optional[List[types_schema.AppID]] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        app_versions: Optional[List[types_schema.AppVersion]] = None,
        run_name: Optional[types_schema.RunName] = None,
        record_ids: Optional[List[types_schema.RecordID]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Tuple[pandas.DataFrame, List[str]]:
        """Get records, their feedback results, and feedback names.

        Args:
            app_ids:
                A list of app ids to filter records by. If empty or not given,
                all apps' records will be returned.
            app_name:
                A name of the app to filter records by. If given, only records
                for this app will be returned.
            app_version:
                A version of the app to filter records by. If given, only
                records for this app version will be returned.
            app_versions:
                A list of app versions to filter records by. If given, only
                records for these app versions will be returned.
            run_name:
                A run name to filter records by. If given, only records for
                this run will be returned.
            record_ids: An optional list of record ids to filter records by.
            offset: Record row offset.
            limit: Limit on the number of records to return.

        Returns:
            Tuple of:
            - DataFrame of records with their feedback results.
            - List of feedback names that are columns in the DataFrame.
        """
        return self.connector.get_records_and_feedback(
            app_ids=app_ids,
            app_name=app_name,
            app_version=app_version,
            app_versions=app_versions,
            run_name=run_name,
            record_ids=record_ids,
            offset=offset,
            limit=limit,
        )

    def get_leaderboard(
        self,
        app_ids: Optional[List[types_schema.AppID]] = None,
        group_by_metadata_key: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> pandas.DataFrame:
        """Get a leaderboard for the given apps.

        Args:
            app_ids: A list of app ids to filter records by. If empty or not given, all
                apps will be included in leaderboard.

            group_by_metadata_key: A key included in record metadata that you want to group results by.

            limit: Limit on the number of records to aggregate to produce the leaderboard.

            offset: Record row offset to select which records to use to aggregate the leaderboard.

        Returns:
            Dataframe of apps with their feedback results aggregated.
            If group_by_metadata_key is provided, the dataframe will be grouped by the specified key.
        """
        return self.connector.get_leaderboard(
            app_ids=app_ids,
            group_by_metadata_key=group_by_metadata_key,
            limit=limit,
            offset=offset,
        )

    def add_ground_truth_to_dataset(
        self,
        dataset_name: str,
        ground_truth_df: pandas.DataFrame,
        dataset_metadata: Optional[Dict[str, Any]] = None,
    ):
        """Create a new dataset, if not existing, and add ground truth data to it. If
        the dataset with the same name already exists, the ground truth data will be added to it.

        Args:
            dataset_name: Name of the dataset.
            ground_truth_df: DataFrame containing the ground truth data.
            dataset_metadata: Additional metadata to add to the dataset.
        """

        # Create and insert the dataset record
        dataset = dataset_schema.Dataset(
            name=dataset_name,
            meta=dataset_metadata,
        )
        dataset_id = self.connector.db.insert_dataset(dataset=dataset)

        buffer = []

        for _, row in ground_truth_df.iterrows():
            ground_truth = groundtruth_schema.GroundTruth(
                dataset_id=dataset_id,
                query=row["query"],
                query_id=row.get("query_id", None),
                expected_response=row.get("expected_response", None),
                expected_chunks=row.get("expected_chunks", None),
                meta=row.get("meta", None),
            )
            buffer.append(ground_truth)

            if len(buffer) >= self.GROUND_TRUTHS_BATCH_SIZE:
                self.connector.db.batch_insert_ground_truth(buffer)
                buffer.clear()

        # remaining ground truths in the buffer
        if buffer:
            self.connector.db.batch_insert_ground_truth(buffer)

    def get_ground_truth(
        self,
        dataset_name: Optional[str] = None,
        user_table_name: Optional[str] = None,
        user_schema_mapping: Optional[Dict[str, str]] = None,
        user_schema_name: Optional[str] = None,
    ) -> pandas.DataFrame:
        """Get ground truth data from the dataset. If `user_table_name` and `user_schema_mapping` are provided,
        load a virtual dataset from the user's table using the schema mapping. If `dataset_name` is provided,
        load ground truth data from the dataset by name.
        dataset_name: Name of the dataset.
        user_table_name: Name of the user's table to load ground truth data from.
        user_schema_mapping: Mapping of user table columns to internal `GroundTruth` schema fields.
        user_schema_name: Name of the user's schema to load ground truth data from.
        """
        if user_table_name and user_schema_mapping:
            return self.connector.db.get_virtual_ground_truth(
                user_table_name, user_schema_mapping, user_schema_name
            )
        elif dataset_name:
            return self.connector.db.get_ground_truths_by_dataset(dataset_name)
        else:
            raise ValueError(
                "Either `dataset_name` or `user_table_name` and `user_schema_mapping` must be provided."
            )

    def start_evaluator(
        self,
        restart: bool = False,
        fork: bool = False,
        disable_tqdm: bool = False,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
        return_when_done: bool = False,
    ) -> Optional[Union[Process, Thread]]:
        """
        Start a deferred feedback function evaluation thread or process.

        Args:
            restart: If set, will stop the existing evaluator before starting a
                new one.

            fork: If set, will start the evaluator in a new process instead of a
                thread. NOT CURRENTLY SUPPORTED.

            disable_tqdm: If set, will disable progress bar logging from the evaluator.

            run_location: Run only the evaluations corresponding to run_location.

            return_when_done: Instead of running asynchronously, will block until no feedbacks remain.

        Returns:
            If return_when_done is True, then returns None. Otherwise, the started process or thread
                that is executing the deferred feedback evaluator.

        Relevant constants:
            [RETRY_RUNNING_SECONDS][trulens.core.session.TruSession.RETRY_RUNNING_SECONDS]

            [RETRY_FAILED_SECONDS][trulens.core.session.TruSession.RETRY_FAILED_SECONDS]

            [DEFERRED_NUM_RUNS][trulens.core.session.TruSession.DEFERRED_NUM_RUNS]

            [MAX_THREADS][trulens.core.utils.threading.TP.MAX_THREADS]
        """
        if is_otel_tracing_enabled():
            raise ValueError(
                "Deferred evaluator not supported with OTEL tracing!"
            )

        assert not fork, "Fork mode not yet implemented."
        assert (
            (not fork) or (not return_when_done)
        ), "fork=True implies running asynchronously but return_when_done=True does not!"

        if self._evaluator_proc is not None:
            if restart:
                self.stop_evaluator()
            else:
                raise RuntimeError(
                    "Evaluator is already running in this process."
                )

        if not fork:
            self._evaluator_stop = threading.Event()

        def runloop(stop_when_none_left: bool = False):
            assert self._evaluator_stop is not None

            print(
                f"Will keep max of "
                f"{self.DEFERRED_NUM_RUNS} feedback(s) running."
            )
            print(
                f"Tasks are spread among max of "
                f"{threading_utils.TP.MAX_THREADS} thread(s)."
            )
            print(
                f"Will rerun running feedbacks after "
                f"{text_utils.format_seconds(self.RETRY_RUNNING_SECONDS)}."
            )
            print(
                f"Will rerun failed feedbacks after "
                f"{text_utils.format_seconds(self.RETRY_FAILED_SECONDS)}."
            )

            total = 0

            # TODO: a lot of the time we say `if tqdm`, but shouldn't we say `if not disable_tqdm`?
            if tqdm:
                # Getting total counts from the database to start off the tqdm
                # progress bar initial values so that they offer accurate
                # predictions initially after restarting the process.
                queue_stats = self.connector.db.get_feedback_count_by_status()
                queue_done = (
                    queue_stats.get(feedback_schema.FeedbackResultStatus.DONE)
                    or 0
                )
                queue_total = sum(queue_stats.values())

                # Show the overall counts from the database, not just what has been
                # looked at so far.
                tqdm_status = tqdm(
                    desc="Feedback Status",
                    initial=queue_done,
                    unit="feedbacks",
                    total=queue_total,
                    postfix={
                        status.name: count
                        for status, count in queue_stats.items()
                    },
                    disable=disable_tqdm,
                )

                # Show the status of the results so far.
                tqdm_total = tqdm(
                    desc="Done Runs",
                    initial=0,
                    unit="runs",
                    disable=disable_tqdm,
                )

                # Show what is being waited for right now.
                tqdm_waiting = tqdm(
                    desc="Waiting for Runs",
                    initial=0,
                    unit="runs",
                    disable=disable_tqdm,
                )

            runs_stats = defaultdict(int)

            futures_map: Dict[
                Future[feedback_schema.FeedbackResult], pandas.Series
            ] = dict()

            while fork or not self._evaluator_stop.is_set():
                if len(futures_map) < self.DEFERRED_NUM_RUNS:
                    # Get some new evals to run if some already completed by now.
                    new_futures: List[
                        Tuple[
                            pandas.Series,
                            Future[feedback_schema.FeedbackResult],
                        ]
                    ] = core_feedback.Feedback.evaluate_deferred(
                        limit=self.DEFERRED_NUM_RUNS - len(futures_map),
                        shuffle=True,
                        session=self,
                        run_location=run_location,
                    )

                    # Will likely get some of the same ones that already have running.
                    for row, fut in new_futures:
                        if fut in futures_map:
                            # If the future is already in our set, check whether
                            # its status has changed and if so, note it in the
                            # runs_stats.
                            if futures_map[fut].status != row.status:
                                runs_stats[row.status.name] += 1

                        futures_map[fut] = row
                        total += 1

                    if tqdm:
                        tqdm_total.total = total
                        tqdm_total.refresh()

                if tqdm:
                    tqdm_waiting.total = self.DEFERRED_NUM_RUNS
                    tqdm_waiting.n = len(futures_map)
                    tqdm_waiting.refresh()

                # Note whether we have waited for some futures in this
                # iteration. Will control some extra wait time if there is no
                # work.
                did_wait = False

                if len(futures_map) > 0:
                    did_wait = True

                    futures_copy = list(futures_map.keys())

                    try:
                        for fut in futures.as_completed(
                            futures_copy, timeout=10
                        ):
                            del futures_map[fut]

                            if tqdm:
                                tqdm_waiting.update(-1)
                                tqdm_total.update(1)

                            feedback_result = fut.result()
                            runs_stats[feedback_result.status.name] += 1

                    except futures.TimeoutError:
                        pass

                if tqdm:
                    tqdm_total.set_postfix({
                        name: count for name, count in runs_stats.items()
                    })

                    queue_stats = (
                        self.connector.db.get_feedback_count_by_status()
                    )
                    queue_done = (
                        queue_stats.get(
                            feedback_schema.FeedbackResultStatus.DONE
                        )
                        or 0
                    )
                    queue_total = sum(queue_stats.values())

                    tqdm_status.n = queue_done
                    tqdm_status.total = queue_total
                    tqdm_status.set_postfix({
                        status.name: count
                        for status, count in queue_stats.items()
                    })

                # Check if any of the running futures should be stopped.
                futures_copy = list(futures_map.keys())
                for fut in futures_copy:
                    row = futures_map[fut]

                    if fut.running():
                        # Not checking status here as this will be not yet be set
                        # correctly. The computation in the future updates the
                        # database but this object is outdated.

                        elapsed = datetime.now().timestamp() - row.last_ts
                        if elapsed > self.RETRY_RUNNING_SECONDS:
                            fut.cancel()

                            # Not an actual status, but would be nice to
                            # indicate cancellations in run stats:
                            runs_stats["CANCELLED"] += 1

                            del futures_map[fut]

                if not did_wait:
                    if stop_when_none_left:
                        break
                    # Nothing to run/is running, wait a bit.
                    if fork:
                        sleep(10)
                    else:
                        self._evaluator_stop.wait(10)

            print("Evaluator stopped.")

        if return_when_done:
            runloop(stop_when_none_left=True)
            return None
        else:
            if fork:
                proc = Process(target=runloop)
            else:
                proc = threading_utils.Thread(target=runloop)
                proc.daemon = True
            # Start a persistent thread or process that evaluates feedback functions.
            self._evaluator_proc = proc
            proc.start()
            return proc

    run_evaluator = start_evaluator

    def stop_evaluator(self):
        """
        Stop the deferred feedback evaluation thread.
        """

        if self._evaluator_proc is None:
            raise RuntimeError("Evaluator not running this process.")

        if isinstance(self._evaluator_proc, Process):
            self._evaluator_proc.terminate()

        elif isinstance(self._evaluator_proc, Thread):
            self._evaluator_stop.set()
            self._evaluator_proc.join()
            self._evaluator_stop = None

        self._evaluator_proc = None

    def wait_for_records(
        self,
        record_ids: List[str],
        timeout: float = 10,
        poll_interval: float = 0.5,
    ) -> None:
        """
        Wait for specific record_ids to appear in the TruLens session.

        Args:
            record_ids: The record ids to wait for.
            timeout: Maximum time to wait in seconds.
            poll_interval: How often to poll in seconds.
        """
        if is_otel_tracing_enabled():
            self.force_flush()
            self._wait_for_records_otel(record_ids, timeout, poll_interval)
        else:
            self._wait_for_records_legacy(record_ids, timeout, poll_interval)

    def _wait_for_records_otel(
        self,
        record_ids: List[str],
        timeout: float,
        poll_interval: float,
    ) -> None:
        """OTEL-specific implementation to wait for records.

        This is more efficient than the legacy implementation as it only checks
        for the existence of RECORD_ROOT spans rather than fetching all record
        data and feedback results.
        """
        start_time = time()
        while time() - start_time < timeout:
            # Get events for the specific record_ids without any start_time filter
            events = self.connector.get_events(
                app_name=None,
                app_version=None,
                record_ids=record_ids,
                start_time=None,
            )

            if events is not None and not events.empty:
                # Check for RECORD_ROOT spans to confirm complete records
                found_record_ids = set()
                for _, event in events.iterrows():
                    record_attrs = event.get("record_attributes", {})
                    if isinstance(record_attrs, str):
                        record_attrs = json.loads(record_attrs)
                    span_type = record_attrs.get(SpanAttributes.SPAN_TYPE)
                    record_id = record_attrs.get(SpanAttributes.RECORD_ID)
                    if (
                        span_type == SpanAttributes.SpanType.RECORD_ROOT
                        and record_id is not None
                    ):
                        found_record_ids.add(record_id)

                if all(
                    record_id in found_record_ids for record_id in record_ids
                ):
                    return

            sleep(poll_interval)

        raise RuntimeError(
            f"Could not find all record IDs: {record_ids} in database!"
        )

    def _wait_for_records_legacy(
        self,
        record_ids: List[str],
        timeout: float,
        poll_interval: float,
    ) -> None:
        """Legacy (pre-OTEL) implementation to wait for records."""
        start_time = time()
        while time() - start_time < timeout:
            records_df, _ = self.get_records_and_feedback(
                record_ids=record_ids,
            )
            known_record_ids = set(records_df["record_id"])
            if not records_df.empty and all(
                record_id in known_record_ids for record_id in record_ids
            ):
                return
            sleep(poll_interval)
        raise RuntimeError(
            f"Could not find all record IDs: {record_ids} in database!"
        )

    def wait_for_feedback_results(
        self,
        record_ids: List[str],
        feedback_names: List[str],
        timeout: float = 60,
        poll_interval: float = 0.5,
    ) -> None:
        """
        Wait for feedback evaluation results to appear in the database.

        This method waits for EVAL_ROOT spans to appear for all combinations of
        the given record_ids and feedback_names. It should be called after
        feedback computation has been triggered (e.g., after compute_now).

        Args:
            record_ids: The record IDs to wait for feedback results for.
            feedback_names: The names of the feedback functions to wait for.
            timeout: Maximum time to wait in seconds.
            poll_interval: How often to poll in seconds.

        Raises:
            RuntimeError: If not all feedback results are found within the timeout.
        """
        if not is_otel_tracing_enabled():
            # In legacy mode, feedback results are written synchronously
            # so no waiting is needed
            return

        self.force_flush()

        # Build the set of expected (record_id, feedback_name) pairs
        expected_pairs = {
            (record_id, feedback_name)
            for record_id in record_ids
            for feedback_name in feedback_names
        }

        found_pairs = set()
        start_time = time()
        while time() - start_time < timeout:
            # Get events for the specific record_ids without any start_time filter
            events = self.connector.get_events(
                app_name=None,
                app_version=None,
                record_ids=record_ids,
                start_time=None,
            )

            if events is not None and not events.empty:
                # Check for EVAL_ROOT spans and collect found (record_id, metric_name) pairs
                found_pairs = set()
                for _, event in events.iterrows():
                    record_attrs = event.get("record_attributes", {})
                    if isinstance(record_attrs, str):
                        record_attrs = json.loads(record_attrs)

                    span_type = record_attrs.get(SpanAttributes.SPAN_TYPE)
                    if span_type != SpanAttributes.SpanType.EVAL_ROOT:
                        continue

                    record_id = record_attrs.get(SpanAttributes.RECORD_ID)
                    metric_name = record_attrs.get(
                        SpanAttributes.EVAL_ROOT.METRIC_NAME
                    )

                    if record_id is not None and metric_name is not None:
                        found_pairs.add((record_id, metric_name))

                # Check if all expected pairs have been found
                if expected_pairs.issubset(found_pairs):
                    return

            sleep(poll_interval)

        # Build a helpful error message showing what's missing
        missing_pairs = expected_pairs - found_pairs
        raise RuntimeError(
            f"Timeout waiting for feedback results. Missing: {missing_pairs}"
        )

    def add_feedback_result(
        self,
        record: Record,
        feedback_name: str,
        feedback_result: Union[float, int],
        higher_is_better: bool,
    ) -> None:
        """
        Add a feedback result for a given record.

        Args:
            record: The Record object to add feedback for.
            feedback_name: The name of the feedback function.
            feedback_result: The feedback score/result (float or int).
            higher_is_better: Whether higher values are better.
        """
        if not is_otel_tracing_enabled():
            raise RuntimeError(
                "add_feedback_result is only supported when OTEL tracing is enabled!"
            )

        from trulens.core.feedback.feedback import Feedback
        from trulens.feedback.computer import _call_feedback_function

        # Get necessary information from the record root.
        record_root_event = record._get_record_root_event()
        resource_attributes = record_root_event["resource_attributes"]
        record_attributes = record_root_event["record_attributes"]
        app_name = resource_attributes.get(ResourceAttributes.APP_NAME)
        app_version = resource_attributes.get(ResourceAttributes.APP_VERSION)
        app_id = resource_attributes.get(ResourceAttributes.APP_ID)
        run_name = record_attributes.get(SpanAttributes.RUN_NAME)
        input_id = record_attributes.get(SpanAttributes.INPUT_ID)

        # Wrap constant result in a Feedback object to align with new API
        const_feedback = Feedback(
            lambda: feedback_result,
            name=feedback_name,
            higher_is_better=higher_is_better,
        )

        # Create feedback computation recording context
        feedback_result = _call_feedback_function(
            feedback_name,
            const_feedback,
            higher_is_better,
            None,
            {},
            app_name,
            app_version,
            app_id,
            run_name,
            input_id,
            record.record_id,
            None,
        )

    def compute_feedbacks_on_events(
        self,
        events: pandas.DataFrame,
        feedbacks: List[core_metric.Metric],
        raise_error_on_no_feedbacks_computed: bool = False,
    ) -> None:
        """Compute feedbacks/metrics on events.

        Args:
            events:
                Events to compute feedbacks on. This can be from multiple
                records.
            feedbacks: Feedback functions to compute.
            raise_error_on_no_feedbacks_computed:
                Raise an error if no feedbacks were computed. Default is False.
        """
        if not is_otel_tracing_enabled():
            raise ValueError(
                "This method is only supported for OTEL Tracing. Please enable OTEL tracing in the environment!"
            )

        try:
            from trulens.feedback.computer import compute_feedback_by_span_group
        except ImportError:
            logger.error(
                "trulens.feedback package is not installed. Please install it to use feedback computation functionality."
            )
            raise

        for feedback in feedbacks:
            compute_feedback_by_span_group(
                events,
                feedback,
                raise_error_on_no_feedbacks_computed,
            )

    def get_events(
        self,
        app_name: Optional[str],
        app_version: Optional[str],
        record_ids: Optional[List[str]] = None,
        start_time: Optional[datetime] = None,
    ) -> pandas.DataFrame:
        """
        Get events/spans from the database in OTel mode.

        Args:
            app_name: The app name to filter events by.
            app_version: The app version to filter events by.
            record_ids: The record ids to filter events by.
            start_time: The minimum time to consider events from.

        Returns:
            A pandas DataFrame of all relevant events/spans.
        """
        return self.connector.get_events(
            app_name=app_name,
            app_version=app_version,
            record_ids=record_ids,
            start_time=start_time,
        )


def Tru(*args, **kwargs) -> TruSession:
    warnings.warn(
        "Tru is deprecated, use TruSession instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return TruSession(*args, **kwargs)



"""
This is pre-sqlalchemy db migration. This file should not need changes. It is
here for backwards compatibility of oldest TruLens versions.
"""

import json
import logging
import shutil
import traceback
from typing import Callable, List
import uuid

import pydantic
from trulens.core.schema import app as app_schema
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.utils import pyschema as pyschema_utils

logger = logging.getLogger(__name__)
"""
How to make a db migrations:

1. Create a compatibility DB (checkout the last pypi rc branch https://github.com/truera/trulens/tree/releases/rc-trulens-X.x.x/):
  In trulens/tests/docs_notebooks/notebooks_to_test
  remove any local dbs
    * rm rf default.sqlite
  run below notebooks (Making sure you also run with the same X.x.x version trulens)
    * all_tools.ipynb # cp cp ../generated_files/all_tools.ipynb ./
    * llama_index_quickstart.ipynb # cp frameworks/llama_index/llama_index_quickstart.ipynb ./
    * langchain-retrieval-augmentation-with-trulens.ipynb # cp vector-dbs/pinecone/langchain-retrieval-augmentation-with-trulens.ipynb ./
    * Add any other notebooks you think may have possible breaking changes
  replace the last compatible db with this new db file
    * See the last COMPAT_VERSION: compatible version in leftmost below: migration_versions
    * mv default.sqlite trulens/release_dbs/COMPAT_VERSION/default.sqlite

2. Do Migration coding
  * Update __init__.py with the new version
  * The upgrade methodology is determined by this data structure
        upgrade_paths = {
            # from_version: (to_version,migrate_function)
            "0.1.2": ("0.2.0", migrate_0_1_2),
            "0.2.0": ("0.3.0", migrate_0_2_0)
        }
  * add your version to the version list:
      migration_versions: list = [YOUR VERSION HERE,...,"0.3.0", "0.2.0", "0.1.2"]


3. To Test
  * replace your db file with an old version db first and see if the session.migrate_database() works.

4. Add a DB file for testing new breaking changes (Same as step 1: but with your new version)
  * Do a sys.path.insert(0,TRULENS_PATH) to run with your version
"""


class VersionException(Exception):
    pass


MIGRATION_UNKNOWN_STR = "unknown[db_migration]"
migration_versions: List[str] = ["1.4.9"]


def _update_db_json_col(
    db, table: str, old_entry: tuple, json_db_col_idx: int, new_json: dict
):
    """Replaces an old json serialized db column with a migrated/new one.

    Args:
        db (DB): the db object

        table (str): the table to update (from the current DB)

        old_entry (tuple): the db tuple to update

        json_db_col_idx (int): the tuple idx to update

        new_json (dict): the new json object to be put in the D
    """
    migrate_record = list(old_entry)
    migrate_record[json_db_col_idx] = json.dumps(new_json)
    migrate_record = tuple(migrate_record)
    db._insert_or_replace_vals(table=table, vals=migrate_record)


def jsonlike_map(fval=None, fkey=None, fkeyval=None):
    if fval is None:
        fval = lambda x: x
    if fkey is None:
        fkey = lambda x: x
    if fkeyval is None:
        fkeyval = lambda x, y: (x, y)

    def walk(obj):
        if isinstance(obj, dict):
            ret = {}
            for k, v in obj.items():
                k = fkey(k)
                v = fval(walk(v))
                k, v = fkeyval(k, v)
                ret[k] = v
            return fval(ret)

        if isinstance(obj, (list, tuple)):
            return fval(type(obj)(fval(walk(v)) for v in obj))

        else:
            return fval(obj)

    return walk


def jsonlike_rename_key(old_key, new_key) -> Callable:
    def fkey(k):
        if k == old_key:
            logger.debug(f"key {old_key} -> {new_key}")
            return new_key
        else:
            return k

    return jsonlike_map(fkey=fkey)


def jsonlike_rename_value(old_val, new_val) -> Callable:
    def fval(v):
        if v == old_val:
            logger.debug(f"value {old_val} -> {new_val}")
            return new_val
        else:
            return v

    return jsonlike_map(fval=fval)


class UnknownClass(pydantic.BaseModel):
    def unknown_method(self):
        """
        This is a placeholder put into the database in place of methods whose
        information was not recorded in earlier versions of trulens.
        """


upgrade_paths = {
    # "from_version":("to_version", migrate_method)
    # "0.9.0": ("0.19.0", migrate_0_9_0)
}


def _parse_version(version_str: str) -> List[str]:
    """
    Takes a version string and returns a list of major, minor, patch.

    Args:
        - version_str (str): a version string

    Returns:
        list: [major, minor, patch] strings
    """
    return version_str.split(".")


def _get_compatibility_version(version: str) -> str:
    """
    Gets the db version that the pypi version is compatible with.

    Args:
        - version (str): a pypi version

    Returns:
        - str: a backwards compat db version
    """

    version_split = _parse_version(version)

    for m_version_str in migration_versions:
        for i, m_version_split in enumerate(_parse_version(m_version_str)):
            if int(version_split[i]) > int(m_version_split):
                return m_version_str

            elif int(version_split[i]) == int(m_version_split):
                if i == 2:  # patch version
                    return m_version_str
                # Can't make a choice here, move to next endian.
                continue

            else:
                # The m_version from m_version_str is larger than this version
                # check the next m_version.
                break
    raise VersionException(
        f"Could not find a compatibility version for version {version}"
    )


def _migration_checker(db, warn: bool = False) -> None:
    """
    Checks whether this db, if pre-populated, is compatible with this pypi
    version.

    Args:
        - db (DB): the db object to check
        - warn (bool, optional): if warn is False, then a migration issue will
          raise an exception, otherwise allow passing but only warn. Defaults to
          False.
    """
    meta = db.get_meta()

    _check_needs_migration(meta.trulens_version, warn=warn)


def commit_migrated_version(db, version: str) -> None:
    """After a successful migration, update the DB meta version

    Args:
        db (DB): the db object
        version (str): The version string to set this DB to
    """
    conn, c = db._connect()

    c.execute(
        f"""UPDATE {db.TABLE_META}
                SET value = '{version}'
                WHERE key='trulens_version';
            """
    )
    conn.commit()


def _upgrade_possible(compat_version: str) -> bool:
    """
    Checks the upgrade paths to see if there is a valid migration from the DB to
    the current pypi version

    Args:
        - compat_version (str): the current db version.

    Returns:
        - bool: True if there is an upgrade path. False if not.
    """
    while compat_version in upgrade_paths:
        compat_version = upgrade_paths[compat_version][0]
    return compat_version == migration_versions[0]


def _check_needs_migration(version: str, warn=False) -> None:
    """
    Checks whether the from DB version can be updated to the current DB version.

    Args:
        - version (str): the pypi version
        - warn (bool, optional): if warn is False, then a migration issue will
          raise an exception, otherwise allow passing but only warn. Defaults to
          False.
    """
    compat_version = _get_compatibility_version(version)

    if migration_versions.index(compat_version) > 0:
        if _upgrade_possible(compat_version):
            msg = (
                f"Detected that your db version {version} is from an older release that is incompatible with this release. "
                f"You can either reset your db with `TruSession().reset_database()`, "
                f"or you can initiate a db migration with `TruSession().migrate_database()`"
            )
        else:
            msg = (
                f"Detected that your db version {version} is from an older release that is incompatible with this release and cannot be migrated. "
                f"Reset your db with `TruSession().reset_database()`"
            )
        if warn:
            print(f"Warning! {msg}")
        else:
            raise VersionException(msg)


saved_db_locations = {}


def _serialization_asserts(db) -> None:
    """
    After a successful migration, Do some checks if serialized jsons are loading
    properly.

    Args:
        db (DB): the db object
    """
    global saved_db_locations
    _, c = db._connect()
    SAVED_DB_FILE_LOC = saved_db_locations[db.filename]
    validation_fail_advice = (
        f"Please open a ticket on trulens github page including details on the old and new trulens versions. "
        f"The migration completed so you can still proceed; but stability is not guaranteed. "
        f"Your original DB file is saved here: {SAVED_DB_FILE_LOC} and can be used with the previous version, or you can `TruSession.reset_database()`"
    )

    for table in db.TABLES:
        c.execute(
            f"""PRAGMA table_info({table});
                """
        )
        columns = c.fetchall()
        for col_idx, col in enumerate(columns):
            col_name_idx = 1
            col_name = col[col_name_idx]
            # This is naive for now...
            if "json" in col_name:
                c.execute(f"""SELECT * FROM {table}""")
                rows = c.fetchall()
                for row in rows:
                    try:
                        if row[col_idx] == MIGRATION_UNKNOWN_STR:
                            continue

                        test_json = json.loads(row[col_idx])
                        # special implementation checks for serialized classes
                        if "implementation" in test_json:
                            try:
                                pyschema_utils.FunctionOrMethod.model_validate(
                                    test_json["implementation"]
                                ).load()
                            except ImportError:
                                # Import error is not a migration problem.
                                # It signals that the function cannot be used for deferred evaluation.
                                pass

                        if col_name == "record_json":
                            record_schema.Record.model_validate(test_json)
                        elif col_name == "cost_json":
                            base_schema.Cost.model_validate(test_json)
                        elif col_name == "perf_json":
                            base_schema.Perf.model_validate(test_json)
                        elif col_name == "calls_json":
                            for record_app_call_json in test_json["calls"]:
                                feedback_schema.FeedbackCall.model_validate(
                                    record_app_call_json
                                )
                        elif col_name == "feedback_json":
                            feedback_schema.FeedbackDefinition.model_validate(
                                test_json
                            )
                        elif col_name == "app_json":
                            app_schema.AppDefinition.model_validate(test_json)
                        else:
                            # If this happens, trulens needs to add a migration

                            raise VersionException(
                                f"serialized column migration not implemented: {col_name}. {validation_fail_advice}"
                            )
                    except Exception as e:
                        tb = traceback.format_exc()

                        raise VersionException(
                            f"Migration failed on {table} {col_name} {row[col_idx]}.\n\n{tb}\n\n{validation_fail_advice}"
                        ) from e


def migrate(db) -> None:
    """Migrate a db to the compatible version of this pypi version

    Args:
        db (DB): the db object
    """
    # NOTE TO DEVELOPER: If this method fails: It's likely you made a db
    # breaking change. Follow these steps to add a compatibility change
    # - Update the __init__ version to the next one (if not already)
    # - In this file: add that version to `migration_versions` variable`
    # - Add the migration step in `upgrade_paths` of the form
    #   `from_version`:(`to_version_you_just_created`, `migration_function`)
    # - AFTER YOU PASS TESTS - add your newest db into
    #   `release_dbs/<version_you_just_created>/default.sqlite`
    #   - This is created by running the all_tools and llama_quickstart from a
    #     fresh db (you can `rm -rf` the sqlite file )
    #   - TODO: automate this step
    original_db_file = db.filename
    global saved_db_locations

    saved_db_file = (
        original_db_file.parent
        / f"{original_db_file.name}_saved_{uuid.uuid1()}"
    )
    saved_db_locations[original_db_file] = saved_db_file
    shutil.copy(original_db_file, saved_db_file)
    print(
        f"Saved original db file: `{original_db_file}` to new file: `{saved_db_file}`"
    )

    version = db.get_meta().trulens_version
    from_compat_version = _get_compatibility_version(version)
    while from_compat_version in upgrade_paths:
        to_compat_version, migrate_fn = upgrade_paths[from_compat_version]
        migrate_fn(db=db)
        commit_migrated_version(db=db, version=to_compat_version)
        from_compat_version = to_compat_version

    print("DB Migration complete!")
    _serialization_asserts(db)
    print("DB Validation complete!")



"""
DEPRECATED: This module is deprecated. Use trulens.core.metric instead.

The Feedback class is now a deprecated alias for Metric.
This module is maintained for backward compatibility.
"""

from __future__ import annotations

import logging
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
)
import warnings

from trulens.core.metric.metric import GroundednessConfigs
from trulens.core.metric.metric import InvalidSelector
from trulens.core.metric.metric import Metric
from trulens.core.metric.metric import SkipEval
from trulens.core.schema import feedback as feedback_schema

logger = logging.getLogger(__name__)

# Re-export these for backward compatibility
__all__ = [
    "Feedback",
    "SnowflakeFeedback",
    "SkipEval",
    "InvalidSelector",
    "GroundednessConfigs",
]


class Feedback(Metric):
    """
    DEPRECATED: Use Metric instead.

    This class is maintained for backward compatibility. All functionality
    has been moved to the Metric class.

    Example of migrating to Metric:
        ```python
        # Old way (deprecated):
        from trulens.core import Feedback
        feedback = Feedback(provider.relevance).on_input_output()

        # New way (recommended):
        from trulens.core import Metric
        metric = Metric(implementation=provider.relevance).on_input_output()
        ```
    """

    def __init__(
        self,
        imp: Optional[Callable] = None,
        agg: Optional[Callable] = None,
        examples: Optional[List[Tuple]] = None,
        criteria: Optional[str] = None,
        additional_instructions: Optional[str] = None,
        min_score_val: Optional[int] = 0,
        max_score_val: Optional[int] = 3,
        temperature: Optional[float] = 0.0,
        groundedness_configs: Optional[GroundednessConfigs] = None,
        enable_trace_compression: Optional[bool] = None,
        **kwargs,
    ):
        """Initialize a Feedback (deprecated, use Metric instead).

        Args:
            imp: The feedback function to execute. DEPRECATED: Use `implementation` instead.
            agg: Aggregator function for combining multiple feedback results.
            examples: User-supplied examples for this feedback function.
            criteria: Criteria for the feedback evaluation.
            additional_instructions: Custom instructions for the feedback function.
            min_score_val: Minimum score value (default: 0).
            max_score_val: Maximum score value (default: 3).
            temperature: Temperature parameter for LLM-based feedback (default: 0.0).
            groundedness_configs: Optional groundedness configuration.
            enable_trace_compression: Whether to compress trace data.
            **kwargs: Additional arguments passed to parent class.
        """
        warnings.warn(
            "Feedback is deprecated and will be removed in a future version. "
            "Use Metric instead:\n"
            "  from trulens.core import Metric\n"
            "  metric = Metric(implementation=fn).on_input_output()",
            DeprecationWarning,
            stacklevel=2,
        )

        # Pass imp as implementation to Metric
        # Metric's __init__ handles the 'imp' parameter for backward compatibility
        super().__init__(
            imp=imp,
            agg=agg,
            examples=examples,
            criteria=criteria,
            additional_instructions=additional_instructions,
            min_score_val=min_score_val,
            max_score_val=max_score_val,
            temperature=temperature,
            groundedness_configs=groundedness_configs,
            enable_trace_compression=enable_trace_compression,
            **kwargs,
        )


class SnowflakeFeedback(Feedback):
    """[DEPRECATED] Similar to the parent class Feedback except this ensures the feedback is run only on the Snowflake server.

    This class is deprecated and will be removed in the next major release. Please use Metric or [Snowflake AI Observability](https://docs.snowflake.com/en/user-guide/snowflake-cortex/ai-observability/evaluate-ai-applications) instead.
    """

    def __init__(
        self,
        imp: Optional[Callable] = None,
        agg: Optional[Callable] = None,
        **kwargs,
    ):
        # Note: Feedback.__init__ will emit its own deprecation warning
        # We add an additional one specific to SnowflakeFeedback
        warnings.warn(
            "SnowflakeFeedback is deprecated and will be removed in the next major release. "
            "Please use Metric or Snowflake AI Observability instead: "
            "https://docs.snowflake.com/en/user-guide/snowflake-cortex/ai-observability/evaluate-ai-applications",
            DeprecationWarning,
            stacklevel=2,
        )
        if (
            not hasattr(imp, "__self__")
            or str(type(imp.__self__))
            != "<class 'trulens.providers.cortex.provider.Cortex'>"
        ):
            raise ValueError(
                "`SnowflakeFeedback` can only support feedback functions defined in "
                "`trulens-providers-cortex` package's `trulens.providers.cortex.provider.Cortex` class!"
            )
        super().__init__(imp, agg, **kwargs)
        self.run_location = feedback_schema.FeedbackRunLocation.SNOWFLAKE


# Rebuild models to ensure proper pydantic serialization
Feedback.model_rebuild()



"""
USER-FACING METRIC CONTAINER: Runtime wrapper that users interact with to create evaluation metrics.
Handles execution, parameter management (examples, criteria, scoring), and database integration.

This is the new primary API for evaluation metrics, replacing the Feedback class.
"""

from __future__ import annotations

from datetime import datetime
import inspect
from inspect import Signature
from inspect import signature
import itertools
import json
import logging
from pprint import pformat
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
import warnings

import munch
import numpy as np
import pandas
import pydantic
from pydantic import BaseModel
from rich import print as rprint
from rich.markdown import Markdown
from rich.pretty import pretty_repr
from trulens.core._utils import pycompat as pycompat_utils
from trulens.core.feedback import endpoint as core_endpoint
from trulens.core.metric.selector import Selector
from trulens.core.otel.utils import is_otel_tracing_enabled
from trulens.core.schema import app as app_schema
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import select as select_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import deprecation as deprecation_utils
from trulens.core.utils import json as json_utils
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils
from trulens.core.utils import threading as threading_utils

if TYPE_CHECKING:
    from trulens.core import session as core_session


logger = logging.getLogger(__name__)
A = TypeVar("A")

ImpCallable = Callable[[A], Union[float, Tuple[float, Dict[str, Any]]]]
"""Signature of metric implementations.

Those take in any number of arguments and return either a single float or a
float and a dictionary (of metadata)."""

AggCallable = Callable[
    [Union[Iterable[float], Iterable[Tuple[float, float]]]], float
]
"""Signature of aggregation functions."""


class SkipEval(Exception):
    """Raised when evaluating a metric function implementation to skip it so
    it is not aggregated with other non-skipped results.

    Args:
        reason: Optional reason for why this evaluation was skipped.

        metric: The Metric instance this run corresponds to.

        ins: The arguments to this run.
    """

    def __init__(
        self,
        reason: Optional[str] = None,
        metric: Optional[Metric] = None,
        ins: Optional[Dict[str, Any]] = None,
    ):
        self.reason = reason
        self.metric = metric
        self.ins = ins

    def __str__(self):
        return "Metric evaluation skipped" + (
            (" because " + self.reason) if self.reason else ""
        )

    def __repr__(self):
        return f"SkipEval(reason={self.reason})"


class InvalidSelector(Exception):
    """Raised when a selector names something that is missing in a record/app."""

    def __init__(
        self,
        selector: serial_utils.Lens,
        source_data: Optional[Dict[str, Any]] = None,
    ):
        self.selector = selector
        self.source_data = source_data

    def __str__(self):
        return f"Selector {self.selector} does not exist in source data."

    def __repr__(self):
        return f"InvalidSelector({self.selector})"


class GroundednessConfigs(BaseModel):
    use_sent_tokenize: bool
    filter_trivial_statements: bool


class Metric(feedback_schema.FeedbackDefinition):
    """Metric function container.

    A metric evaluates some aspect of an AI application's behavior,
    such as relevance, groundedness, or custom quality measures.

    Example:
        ```python
        from trulens.core import Metric, Selector

        # Direct construction with selectors
        metric = Metric(
            name="relevance_v1",
            implementation=my_relevance_fn,
            selectors={
                "query": Selector.select_record_input(),
                "response": Selector.select_record_output(),
            },
        )

        # Or with a provider and fluent API
        from trulens.providers.openai import OpenAI
        provider = OpenAI()

        metric = Metric(
            implementation=provider.relevance
        ).on_input_output()
        ```

    Note:
        The `enable_trace_compression` parameter is only applicable to metrics
        that take 'trace' as an input parameter. It has no effect on other metrics.
    """

    imp: Optional[ImpCallable] = pydantic.Field(None, exclude=True)
    """Implementation callable.

    A serialized version is stored at
    [FeedbackDefinition.implementation][trulens.core.schema.feedback.FeedbackDefinition.implementation].
    """

    agg: Optional[AggCallable] = pydantic.Field(None, exclude=True)
    """Aggregator method for metrics that produce more than one result.

    A serialized version is stored at
    [FeedbackDefinition.aggregator][trulens.core.schema.feedback.FeedbackDefinition.aggregator].
    """

    examples: Optional[List[Tuple]] = pydantic.Field(None, exclude=True)
    """Examples to use when evaluating the metric."""

    criteria: Optional[str] = pydantic.Field(None, exclude=True)
    """Criteria for the metric."""

    additional_instructions: Optional[str] = pydantic.Field(None, exclude=True)
    """Additional instructions for the metric."""

    min_score_val: Optional[int] = pydantic.Field(None, exclude=True)
    """Minimum score value for the metric."""

    max_score_val: Optional[int] = pydantic.Field(None, exclude=True)
    """Maximum score value for the metric."""

    temperature: Optional[float] = pydantic.Field(None, exclude=True)
    """Temperature parameter for the metric."""

    groundedness_configs: Optional[GroundednessConfigs] = pydantic.Field(
        None, exclude=True
    )
    """Optional groundedness configuration parameters."""

    enable_trace_compression: Optional[bool] = pydantic.Field(
        None, exclude=True
    )
    """Whether to compress trace data to reduce token usage when sending traces to metrics.

    When True, traces are compressed to preserve essential information while removing redundant data.
    When False, full uncompressed traces are used. When None (default), the metric's
    default behavior is used. This flag is only applicable to metrics that take
    'trace' as an input parameter."""

    def __init__(
        self,
        implementation: Optional[Callable] = None,
        agg: Optional[Callable] = None,
        examples: Optional[List[Tuple]] = None,
        criteria: Optional[str] = None,
        additional_instructions: Optional[str] = None,
        min_score_val: Optional[int] = 0,
        max_score_val: Optional[int] = 3,
        temperature: Optional[float] = 0.0,
        groundedness_configs: Optional[GroundednessConfigs] = None,
        enable_trace_compression: Optional[bool] = None,
        metric_type: Optional[str] = None,
        description: Optional[str] = None,
        # Legacy parameter name support
        imp: Optional[Callable] = None,
        **kwargs,
    ):
        """Initialize a metric.

        Args:
            implementation: The metric function to execute. Can be a plain
                callable or a method from a Provider class.
            agg: Aggregator function for combining multiple metric results.
                Defaults to np.mean.
            examples: User-supplied examples for this metric.
            criteria: Criteria for the metric evaluation.
            additional_instructions: Custom instructions for the metric.
            min_score_val: Minimum score value (default: 0).
            max_score_val: Maximum score value (default: 3).
            temperature: Temperature parameter for LLM-based metrics (default: 0.0).
            groundedness_configs: Optional groundedness configuration.
            enable_trace_compression: Whether to compress trace data.
            metric_type: Implementation identifier (e.g., "relevance", "groundedness").
                If not provided, defaults to the function name.
            description: Human-readable description of what this metric measures.
            imp: DEPRECATED. Use `implementation` instead.
            **kwargs: Additional arguments passed to parent class.
        """
        # Handle deprecated parameter name
        additional_instructions = deprecation_utils.handle_deprecated_kwarg(
            kwargs,
            "custom_instructions",
            "additional_instructions",
            additional_instructions,
        )

        # Support legacy 'imp' parameter name - if imp is provided, use it as the callable
        if imp is not None:
            # imp is the actual callable to use
            if implementation is None:
                implementation = imp
            elif isinstance(implementation, pyschema_utils.FunctionOrMethod):
                # Both imp (callable) and implementation (serialized) provided
                # Use imp as callable, keep implementation for storage
                kwargs["implementation"] = implementation
                implementation = imp
            elif isinstance(implementation, dict):
                # Both imp and serialized dict provided - use imp, keep dict for storage
                kwargs["implementation"] = implementation
                implementation = imp
            # else: implementation is also a callable, imp takes precedence

        # imp is the python function/method while implementation is a serialized
        # json structure. Create the one that is missing based on the one that
        # is provided:
        if implementation is not None:
            # Check if implementation is already serialized (dict or FunctionOrMethod) or a callable
            if isinstance(implementation, dict):
                # Raw dict - validate and load it
                kwargs["implementation"] = implementation
                implementation = (
                    pyschema_utils.FunctionOrMethod.model_validate(
                        implementation
                    ).load()
                    if implementation is not None
                    else None
                )
            elif isinstance(implementation, pyschema_utils.FunctionOrMethod):
                # Already a FunctionOrMethod object (Pydantic may have converted it)
                # Keep it in kwargs for parent and load the callable
                kwargs["implementation"] = implementation
                implementation = implementation.load()
            else:
                # It's a callable - serialize it for storage
                if "implementation" not in kwargs:
                    try:
                        kwargs["implementation"] = (
                            pyschema_utils.FunctionOrMethod.of_callable(
                                implementation, loadable=True
                            )
                        )

                    except Exception as e:
                        logger.warning(
                            "Metric implementation %s cannot be serialized: %s "
                            "This may be ok unless you are using the deferred feedback mode.",
                            implementation,
                            e,
                        )

                        kwargs["implementation"] = (
                            pyschema_utils.FunctionOrMethod.of_callable(
                                implementation, loadable=False
                            )
                        )

        else:
            if "implementation" in kwargs:
                impl_val = kwargs["implementation"]
                if isinstance(impl_val, pyschema_utils.FunctionOrMethod):
                    # Already a FunctionOrMethod object - just load it
                    implementation = impl_val.load()
                elif impl_val is not None:
                    # Raw dict - validate and load
                    implementation = (
                        pyschema_utils.FunctionOrMethod.model_validate(
                            impl_val
                        ).load()
                    )

        # Similarly with agg and aggregator.
        if agg is not None:
            # Check if agg is already serialized (dict or FunctionOrMethod) or a callable
            if isinstance(agg, dict):
                # Raw dict - validate and load it
                kwargs["aggregator"] = agg
                agg = pyschema_utils.FunctionOrMethod.model_validate(agg).load()
            elif isinstance(agg, pyschema_utils.FunctionOrMethod):
                # Already a FunctionOrMethod object (Pydantic may have converted it)
                kwargs["aggregator"] = agg
                agg = agg.load()
            elif kwargs.get("aggregator") is None:
                try:
                    # These are for serialization to/from json and for db storage.
                    kwargs["aggregator"] = (
                        pyschema_utils.FunctionOrMethod.of_callable(
                            agg, loadable=True
                        )
                    )
                except Exception as e:
                    # User defined functions in script do not have a module so cannot be serialized
                    logger.warning(
                        "Cannot serialize aggregator %s. "
                        "Deferred mode will default to `np.mean` as aggregator. "
                        "If you are not using `FeedbackMode.DEFERRED`, you can safely ignore this warning. "
                        "%s",
                        agg,
                        e,
                    )
                    # These are for serialization to/from json and for db storage.
                    kwargs["aggregator"] = (
                        pyschema_utils.FunctionOrMethod.of_callable(
                            agg, loadable=False
                        )
                    )

        else:
            agg_val = kwargs.get("aggregator")
            if agg_val is not None:
                if isinstance(agg_val, pyschema_utils.FunctionOrMethod):
                    # Already a FunctionOrMethod object - just load it
                    agg = agg_val.load()
                else:
                    # Raw dict - validate and load
                    agg = pyschema_utils.FunctionOrMethod.model_validate(
                        agg_val
                    ).load()
            else:
                # Default aggregator if neither serialized `aggregator` or
                # loaded `agg` were specified.
                agg = np.mean

        # Pass custom parameters to parent class for serialization
        if criteria is not None:
            kwargs["criteria"] = criteria
        if additional_instructions is not None:
            kwargs["additional_instructions"] = additional_instructions
        if examples is not None:
            kwargs["examples"] = examples
        if metric_type is not None:
            kwargs["metric_type"] = metric_type
        if description is not None:
            kwargs["description"] = description

        super().__init__(**kwargs)

        self.imp = implementation
        self.agg = agg
        self.examples = examples
        self.criteria = criteria
        self.additional_instructions = additional_instructions
        self.min_score_val = min_score_val
        self.max_score_val = max_score_val
        self.temperature = temperature
        self.groundedness_configs = groundedness_configs
        self.enable_trace_compression = enable_trace_compression

        # Verify that `imp` expects the arguments specified in `selectors`:
        if self.imp is not None:
            sig: Signature = signature(self.imp)
            for argname in self.selectors.keys():
                assert argname in sig.parameters, (
                    f"{argname} is not an argument to {self.imp.__name__}. "
                    f"Its arguments are {list(sig.parameters.keys())}."
                )

    def on_input_output(self) -> Metric:
        """
        Specifies that the metric implementation arguments are to be the main
        app input and output in that order.

        Returns a new Metric object with the specification.
        """
        return self.on_input().on_output()

    def on_default(self) -> Metric:
        """
        Specifies that one argument metrics should be evaluated on the main
        app output and two argument metrics should be evaluated on main input
        and main output in that order.

        Returns a new Metric object with this specification.
        """

        if self.imp is None:
            raise ValueError(
                "Metric function implementation is required to determine default argument names."
            )
        sig: Signature = signature(self.imp)
        num_remaining_parameters = len(
            list(k for k in sig.parameters.keys() if k not in self.selectors)
        )
        if num_remaining_parameters == 0:
            return self
        if num_remaining_parameters == 1:
            # If there is only one parameter left, we assume it is the output.
            return self.on_output()
        if num_remaining_parameters == 2:
            # If there are two parameters left, we assume they are the input
            # and output.
            return self.on_input_output()
        # If there are more than two parameters left, we cannot guess what to
        # do.
        raise RuntimeError(
            f"Cannot determine default paths for metric function arguments. "
            f"The metric function has signature {sig}."
        )

    def _print_guessed_selector(self, par_name, par_path):
        if is_otel_tracing_enabled():
            return

        if par_path == select_schema.Select.RecordCalls:
            alias_info = " or `Select.RecordCalls`"
        elif par_path == select_schema.Select.RecordInput:
            alias_info = " or `Select.RecordInput`"
        elif par_path == select_schema.Select.RecordOutput:
            alias_info = " or `Select.RecordOutput`"
        else:
            alias_info = ""

        print(
            f"{text_utils.UNICODE_CHECK} In {self.supplied_name if self.supplied_name is not None else self.name}, "
            f"input {par_name} will be set to {par_path}{alias_info} ."
        )

    @staticmethod
    def evaluate_deferred(
        session: core_session.TruSession,
        limit: Optional[int] = None,
        shuffle: bool = False,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
    ) -> List[
        Tuple[
            pandas.Series,
            pycompat_utils.Future[feedback_schema.FeedbackResult],
        ]
    ]:
        """Evaluates metrics that were specified to be deferred.

        Returns a list of tuples with the DB row containing the Metric and
        initial [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] as
        well as the Future which will contain the actual result.

        Args:
            limit: The maximum number of evals to start.

            shuffle: Shuffle the order of the metrics to evaluate.

            run_location: Only run metrics with this run_location.

        Constants that govern behavior:

        - TruSession.RETRY_RUNNING_SECONDS: How long to time before restarting a metric
          that was started but never failed (or failed without recording that
          fact).

        - TruSession.RETRY_FAILED_SECONDS: How long to wait to retry a failed metric.
        """

        db = session.connector.db

        def prepare_metric(
            row,
        ) -> Optional[feedback_schema.FeedbackResultStatus]:
            record_json = row.record_json
            record = record_schema.Record.model_validate(record_json)

            app_json = row.app_json

            if row.get("feedback_json") is None:
                logger.warning(
                    "Cannot evaluate metric without `feedback_json`. "
                    "This might have come from an old database. \n%s",
                    row,
                )
                return None

            metric = Metric.model_validate(row.feedback_json)

            return metric.run_and_log(
                record=record,
                app=app_json,
                session=session,
                feedback_result_id=row.feedback_result_id,
            )

        # Get the different status metrics except those marked DONE.
        metrics_not_done = db.get_feedback(
            status=[
                feedback_schema.FeedbackResultStatus.NONE,
                feedback_schema.FeedbackResultStatus.FAILED,
                feedback_schema.FeedbackResultStatus.RUNNING,
            ],
            limit=limit,
            shuffle=shuffle,
            run_location=run_location,
        )

        tp = threading_utils.TP()

        futures: List[
            Tuple[
                pandas.Series,
                pycompat_utils.Future[feedback_schema.FeedbackResult],
            ]
        ] = []

        for _, row in metrics_not_done.iterrows():
            now = datetime.now().timestamp()
            elapsed = now - row.last_ts

            if row.status == feedback_schema.FeedbackResultStatus.NONE:
                futures.append((row, tp.submit(prepare_metric, row)))

            elif row.status == feedback_schema.FeedbackResultStatus.RUNNING:
                if elapsed > session.RETRY_RUNNING_SECONDS:
                    futures.append((row, tp.submit(prepare_metric, row)))

                else:
                    pass

            elif row.status == feedback_schema.FeedbackResultStatus.FAILED:
                if elapsed > session.RETRY_FAILED_SECONDS:
                    futures.append((row, tp.submit(prepare_metric, row)))

                else:
                    pass

        return futures

    def __call__(self, *args, **kwargs) -> Any:
        assert (
            self.imp is not None
        ), "Metric definition needs an implementation to call."
        if self.examples is not None:
            kwargs["examples"] = self.examples
        if self.criteria is not None:
            kwargs["criteria"] = self.criteria
        if self.additional_instructions is not None:
            kwargs["additional_instructions"] = self.additional_instructions
        if self.min_score_val is not None:
            kwargs["min_score_val"] = self.min_score_val
        if self.max_score_val is not None:
            kwargs["max_score_val"] = self.max_score_val
        if self.temperature is not None:
            kwargs["temperature"] = self.temperature
        if self.groundedness_configs is not None:
            kwargs["groundedness_configs"] = self.groundedness_configs
        if self.enable_trace_compression is not None:
            kwargs["enable_trace_compression"] = self.enable_trace_compression

        # Filter out unexpected keyword arguments
        sig = signature(self.imp)
        valid_kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}
        return self.imp(*args, **valid_kwargs)

    def aggregate(
        self,
        func: Optional[AggCallable] = None,
        combinations: Optional[feedback_schema.FeedbackCombinations] = None,
    ) -> Metric:
        """
        Specify the aggregation function in case the selectors for this metric
        generate more than one value for implementation argument(s). Can also
        specify the method of producing combinations of values in such cases.

        Returns a new Metric object with the given aggregation function and/or
        the given [combination mode][trulens.core.schema.feedback.FeedbackCombinations].
        """

        if func is None and combinations is None:
            raise ValueError(
                "At least one of `func` or `combinations` must be provided."
            )

        updates = {}
        if func is not None:
            updates["agg"] = func
        if combinations is not None:
            updates["combinations"] = combinations

        return Metric.model_copy(self, update=updates)

    @staticmethod
    def of_feedback_definition(f: feedback_schema.FeedbackDefinition):
        implementation = f.implementation
        aggregator = f.aggregator
        supplied_name = f.supplied_name
        imp_func = implementation.load()
        agg_func = aggregator.load()

        return Metric.model_validate(
            dict(
                imp=imp_func, agg=agg_func, name=supplied_name, **f.model_dump()
            )
        )

    def _next_unselected_arg_name(self):
        if self.imp is not None:
            sig = signature(self.imp)
            par_names = list(
                k for k in sig.parameters.keys() if k not in self.selectors
            )
            if "self" in par_names:
                logger.warning(
                    "Metric function `%s` has `self` as argument. "
                    "Perhaps it is static method or its Provider class was not initialized?",
                    python_utils.callable_name(self.imp),
                )
            if len(par_names) == 0:
                raise TypeError(
                    f"Metric implementation {self.imp} with signature {sig} has no more inputs. "
                    "Perhaps you meant to evaluate it on App output only instead of app input and output?"
                )

            return par_names[0]
        else:
            raise RuntimeError(
                "Cannot determine name of metric function parameter without its definition."
            )

    def on_prompt(self, arg: Optional[str] = None) -> Metric:
        """
        Create a variant of `self` that will take in the main app input or
        "prompt" as input, sending it as an argument `arg` to implementation.
        """
        new_selectors = self.selectors.copy()
        if arg is None:
            arg = self._next_unselected_arg_name()
            self._print_guessed_selector(arg, select_schema.Select.RecordInput)
        if is_otel_tracing_enabled():
            new_selectors[arg] = Selector.select_record_input()
        else:
            new_selectors[arg] = select_schema.Select.RecordInput
        ret = self.model_copy()
        ret.selectors = new_selectors
        return ret

    # alias
    on_input = on_prompt

    def on_response(self, arg: Optional[str] = None) -> Metric:
        """
        Create a variant of `self` that will take in the main app output or
        "response" as input, sending it as an argument `arg` to implementation.
        """
        new_selectors = self.selectors.copy()
        if arg is None:
            arg = self._next_unselected_arg_name()
            self._print_guessed_selector(arg, select_schema.Select.RecordOutput)
        if is_otel_tracing_enabled():
            new_selectors[arg] = Selector.select_record_output()
        else:
            new_selectors[arg] = select_schema.Select.RecordOutput
        ret = self.model_copy()
        ret.selectors = new_selectors
        return ret

    # alias
    on_output = on_response

    def on_context(
        self,
        arg: Optional[str] = None,
        *,
        collect_list: bool,
    ):
        """
        Create a variant of `self` that will attempt to take in the context from
        a context retrieval as input, sending it as an argument `arg` to
        implementation.
        """
        if not is_otel_tracing_enabled():
            raise RuntimeError(
                "Context metrics are only supported in OTel mode!"
            )
        new_selectors = self.selectors.copy()
        if arg is None:
            arg = self._next_unselected_arg_name()
            self._print_guessed_selector(arg, select_schema.Select.RecordOutput)
        new_selectors[arg] = Selector.select_context(collect_list=collect_list)
        ret = self.model_copy()
        ret.selectors = new_selectors
        return ret

    def on(self, *args, **kwargs) -> Metric:
        """
        Create a variant of `self` with the same implementation but the given
        selectors. Those provided positionally get their implementation argument
        name guessed and those provided as kwargs get their name from the kwargs
        key.
        """
        if is_otel_tracing_enabled():
            if len(args) != 1 or len(kwargs) > 0:
                raise ValueError(
                    "OTEL mode only supports a single positional argument to `on`."
                )
            selectors = args[0]
            if not isinstance(selectors, dict):
                raise ValueError(
                    f"OTEL mode only supports dictionary selectors, not {type(selectors)}!"
                )
            sig = signature(self.imp)
            feedback_function_parameters = set(sig.parameters.keys())
            new_selectors = self.selectors.copy()
            for k, v in selectors.items():
                if not isinstance(k, str):
                    raise ValueError(
                        f"OTEL mode only supports string keys, not {type(k)}!"
                    )
                if k not in feedback_function_parameters:
                    raise ValueError(
                        f"Selector key {k} not found in metric function parameters {feedback_function_parameters}!"
                    )
                if not isinstance(v, Selector):
                    raise ValueError(
                        f"OTEL mode only supports Selector values, not {type(v)}!"
                    )
                new_selectors[k] = v
            ret = self.model_copy()
            ret.selectors = new_selectors
            return ret

        new_selectors = self.selectors.copy()

        for k, v in kwargs.items():
            if not isinstance(v, serial_utils.Lens):
                raise ValueError(
                    f"Expected a Lens but got `{v}` of type `{python_utils.class_name(type(v))}`."
                )
            new_selectors[k] = v

        new_selectors.update(kwargs)

        for path in args:
            if not isinstance(path, serial_utils.Lens):
                raise ValueError(
                    f"Expected a Lens but got `{path}` of type `{python_utils.class_name(type(path))}`."
                )

            argname = self._next_unselected_arg_name()
            new_selectors[argname] = path
            self._print_guessed_selector(argname, path)

        ret = self.model_copy()

        ret.selectors = new_selectors

        return ret

    @property
    def sig(self) -> inspect.Signature:
        """Signature of the metric function implementation."""

        if self.imp is None:
            raise RuntimeError(
                "Cannot determine signature of metric function without its definition."
            )

        return signature(self.imp)

    def check_otel_selectors(self):
        if self.imp is None:
            raise RuntimeError(
                "Cannot check selectors of metric function without its definition."
            )

        sig = signature(self.imp)
        function_args = list(sig.parameters.keys())
        required_function_args = [
            param_name
            for param_name, param in sig.parameters.items()
            if param.default == inspect.Parameter.empty
            and param.kind
            not in (
                inspect.Parameter.VAR_KEYWORD,
                inspect.Parameter.VAR_POSITIONAL,
            )
        ]
        error_msg = ""
        # Check for extra selectors. Technically, this shouldn't happen ever
        # since we'd fail before this point, but we check it anyway in case
        # things change.
        extra_selectors = []
        for selector in self.selectors:
            if selector not in function_args:
                extra_selectors.append(selector)
        if extra_selectors:
            error_msg += f"Metric function `{self.name}` has selectors that are not in the function signature:\n"
            error_msg += f"Extra selectors: {extra_selectors}\n"
            error_msg += f"Function args: {function_args}\n"
        # Check for missing selectors.
        missing_selectors = []
        for required_function_arg in required_function_args:
            if required_function_arg not in self.selectors:
                missing_selectors.append(required_function_arg)
        if missing_selectors:
            error_msg += (
                f"Metric function `{self.name}` has missing selectors:\n"
            )
            error_msg += f"Missing selectors: {missing_selectors}\n"
            error_msg += f"Required function args: {required_function_args}\n"
        # Throw error if there are any issues.
        if error_msg:
            raise ValueError(error_msg)

    def check_selectors(
        self,
        app: Union[app_schema.AppDefinition, serial_utils.JSON],
        record: record_schema.Record,
        source_data: Optional[Dict[str, Any]] = None,
        warning: bool = False,
    ) -> bool:
        """Check that the selectors are valid for the given app and record.

        Args:
            app: The app that produced the record.

            record: The record that the metric will run on. This can be a
                mostly empty record for checking ahead of producing one. The
                utility method
                [App.dummy_record][trulens.core.app.App.dummy_record] is built
                for this purpose.

            source_data: Additional data to select from when extracting metric
                function arguments.

            warning: Issue a warning instead of raising an error if a selector is
                invalid. As some parts of a Record cannot be known ahead of
                producing it, it may be necessary to not raise exception here
                and only issue a warning.

        Returns:
            True if the selectors are valid. False if not (if warning is set).

        Raises:
            ValueError: If a selector is invalid and warning is not set.
        """

        from trulens.core.app import App

        if source_data is None:
            source_data = {}

        app_type: str = "trulens recorder (`TruChain`, `TruLlama`, etc)"

        if isinstance(app, App):
            app_type = f"`{type(app).__name__}`"
            app = json_utils.jsonify(
                app,
                instrument=app.instrument,
                skip_specials=True,
                redact_keys=True,
            )

        elif isinstance(app, app_schema.AppDefinition):
            app = json_utils.jsonify(app, skip_specials=True, redact_keys=True)

        source_data = self._construct_source_data(
            app=app, record=record, source_data=source_data
        )

        # Build the hint message here.
        msg = ""

        # Keep track whether any selectors failed to validate.
        check_good: bool = True

        for k, q in self.selectors.items():
            if q.exists(source_data):
                continue

            msg += f"""
# Selector check failed

Source of argument `{k}` to `{self.name}` does not exist in app or expected
record:

```python
{q}
# or equivalently
{select_schema.Select.render_for_dashboard(q)}
```

The data used to make this check may be incomplete. If you expect records
produced by your app to contain the selected content, you can ignore this error
by setting `selectors_nocheck` in the {app_type} constructor. Alternatively,
setting `selectors_check_warning` will print out this message but will not raise
an error.

## Additional information:

Metric function signature:
```python
{self.sig}
```

"""
            prefix = q.existing_prefix(source_data)

            if prefix is None:
                continue

            if (
                len(prefix.path) >= 2
                and isinstance(prefix.path[-1], serial_utils.GetItemOrAttribute)
                and prefix.path[-1].get_item_or_attribute() == "rets"
            ):
                # If the selector check failed because the selector was pointing
                # to something beyond the rets of a record call, we have to
                # ignore it as we cannot tell what will be in the rets ahead of
                # invoking app.
                continue

            if (
                len(prefix.path) >= 3
                and isinstance(prefix.path[-2], serial_utils.GetItemOrAttribute)
                and prefix.path[-2].get_item_or_attribute() == "args"
            ):
                # Likewise if failure was because the selector was pointing to
                # method args beyond their parameter names, we also cannot tell
                # their contents so skip.
                continue

            check_good = False

            msg += f"The prefix `{prefix}` selects this data that exists in your app or typical records:\n\n"

            try:
                for prefix_obj in prefix.get(source_data):
                    if isinstance(prefix_obj, munch.Munch):
                        prefix_obj = prefix_obj.toDict()

                    msg += f"- Object of type `{python_utils.class_name(type(prefix_obj))}` starting with:\n"
                    msg += (
                        "```python\n"
                        + text_utils.retab(
                            tab="\t  ",
                            s=pretty_repr(
                                prefix_obj, max_depth=2, indent_size=2
                            ),
                        )
                        + "\n```\n"
                    )

            except Exception as e:  # pylint: disable=W0718
                msg += f"Some non-existent object because: {pretty_repr(e)}"

        if check_good:
            return True

        # Output using rich text.
        rprint(Markdown(msg))

        if warning:
            return False

        else:
            raise ValueError(
                "Some selectors do not exist in the app or record."
            )

    def run(
        self,
        app: Optional[
            Union[app_schema.AppDefinition, serial_utils.JSON]
        ] = None,
        record: Optional[record_schema.Record] = None,
        source_data: Optional[Dict] = None,
        **kwargs: Dict[str, Any],
    ) -> feedback_schema.FeedbackResult:
        """
        Run the metric on the given `record`. The `app` that
        produced the record is also required to determine input/output argument
        names.

        Args:
            app: The app that produced the record. This can be AppDefinition or a jsonized
                AppDefinition. It will be jsonized if it is not already.

            record: The record to evaluate the metric on.

            source_data: Additional data to select from when extracting metric
                function arguments.

            **kwargs: Any additional keyword arguments are used to set or override
                selected metric function inputs.

        Returns:
            A FeedbackResult object with the result of the metric.
        """

        if isinstance(app, app_schema.AppDefinition):
            app_json = json_utils.jsonify(app)
        else:
            app_json = app

        result_vals = []

        feedback_calls = []

        feedback_result = feedback_schema.FeedbackResult(
            feedback_definition_id=self.feedback_definition_id,
            record_id=record.record_id if record is not None else "no record",
            name=self.supplied_name
            if self.supplied_name is not None
            else self.name,
        )

        source_data = self._construct_source_data(
            app=app_json, record=record, source_data=source_data
        )

        if self.if_exists is not None:
            if not self.if_exists.exists(source_data):
                logger.warning(
                    "Metric %s skipped as %s does not exist.",
                    self.name,
                    self.if_exists,
                )
                feedback_result.status = (
                    feedback_schema.FeedbackResultStatus.SKIPPED
                )
                return feedback_result

        # Separate try block for extracting inputs from records/apps in case a
        # user specified something that does not exist. We want to fail and give
        # a warning earlier than later.
        try:
            input_combinations = list(
                self._extract_selection(
                    source_data=source_data,
                    combinations=self.combinations,
                    **kwargs,
                )
            )

        except InvalidSelector as e:
            # Handle the cases where a selector named something that does not
            # exist in source data.

            if (
                self.if_missing
                == feedback_schema.FeedbackOnMissingParameters.ERROR
            ):
                feedback_result.status = (
                    feedback_schema.FeedbackResultStatus.FAILED
                )
                raise e

            if (
                self.if_missing
                == feedback_schema.FeedbackOnMissingParameters.WARN
            ):
                feedback_result.status = (
                    feedback_schema.FeedbackResultStatus.SKIPPED
                )
                logger.warning(
                    "Metric %s cannot run as %s does not exist in record or app.",
                    self.name,
                    e.selector,
                )
                return feedback_result

            if (
                self.if_missing
                == feedback_schema.FeedbackOnMissingParameters.IGNORE
            ):
                feedback_result.status = (
                    feedback_schema.FeedbackResultStatus.SKIPPED
                )
                return feedback_result

            feedback_result.status = feedback_schema.FeedbackResultStatus.FAILED
            raise ValueError(
                f"Unknown value for `if_missing` {self.if_missing}."
            ) from e

        try:
            # Total cost, will accumulate.
            cost = base_schema.Cost()
            multi_result = None

            # Keep track of evaluations that were skipped due to raising SkipEval.
            skipped_exceptions = []

            for ins in input_combinations:
                try:
                    result_and_meta, get_eval_cost = (
                        core_endpoint.Endpoint.track_all_costs_tally(
                            self, **ins
                        )
                    )

                    cost += get_eval_cost()

                except SkipEval as e:
                    e.metric = self
                    e.ins = ins
                    skipped_exceptions.append(e)
                    warnings.warn(str(e), UserWarning, stacklevel=1)
                    continue  # go to next input_combination

                except Exception as e:
                    raise RuntimeError(
                        f"Evaluation of {self.name} failed on inputs: \n{pformat(ins)[0:128]}."
                    ) from e

                if isinstance(result_and_meta, Tuple):
                    # If output is a tuple of two, we assume it is the float/multifloat and the metadata.
                    assert len(result_and_meta) == 2, (
                        "Metric functions must return either a single float, "
                        "a float-valued dict, or these in combination with a dictionary as a tuple."
                    )
                    result_val, meta = result_and_meta

                    assert isinstance(
                        meta, dict
                    ), f"Metric metadata output must be a dictionary but was {type(meta)}."
                else:
                    # Otherwise it is just the float. We create empty metadata dict.
                    result_val = result_and_meta
                    meta = dict()

                if isinstance(result_val, dict):
                    for val in result_val.values():
                        assert isinstance(val, float), (
                            f"Metric function output with multivalue must be "
                            f"a dict with float values but encountered {type(val)}."
                        )
                    feedback_call = feedback_schema.FeedbackCall(
                        args=ins,
                        ret=np.mean(list(result_val.values())),
                        meta=meta,
                    )

                else:
                    assert isinstance(
                        result_val, (int, float, list, dict)
                    ), f"Metric function output must be a float or an int, a list of floats, or dict but was {type(result_val)}."
                    feedback_call = feedback_schema.FeedbackCall(
                        args=ins, ret=result_val, meta=meta
                    )

                result_vals.append(result_val)
                feedback_calls.append(feedback_call)

            # Warn that there were some skipped evals.
            num_skipped = len(skipped_exceptions)
            num_eval = len(result_vals)
            num_total = num_skipped + num_eval
            if num_skipped > 0 and num_total > 0:
                warnings.warn(
                    (
                        f"{num_skipped}/{num_total}={100.0 * num_skipped / num_total:0.1f}"
                        "% evaluation(s) were skipped because they raised SkipEval "
                        "(see earlier warnings for listing)."
                    ),
                    UserWarning,
                    stacklevel=1,
                )

            if num_eval == 0:
                warnings.warn(
                    f"Metric {self.supplied_name if self.supplied_name is not None else self.name} with aggregation {self.agg} had no inputs.",
                    UserWarning,
                    stacklevel=1,
                )
                result = float("nan")
            else:
                if isinstance(result_vals[0], float):
                    result_vals = np.array(result_vals)
                    result = self.agg(result_vals)
                else:
                    try:
                        # Operates on list of dict; Can be a dict output
                        # (maintain multi) or a float output (convert to single)
                        result = self.agg(result_vals)
                    except Exception:
                        # Alternatively, operate the agg per key
                        result = {}
                        for feedback_output in result_vals:
                            for key in feedback_output:
                                if key not in result:
                                    result[key] = []
                                result[key].append(feedback_output[key])
                        for key in result:
                            result[key] = self.agg(result[key])

                    if isinstance(result, dict):
                        multi_result = result
                        result = np.nan

            feedback_result.update(
                result=result,
                status=feedback_schema.FeedbackResultStatus.DONE,
                cost=cost,
                calls=feedback_calls,
                multi_result=json.dumps(multi_result),
            )

            return feedback_result

        except Exception:
            # Convert traceback to a UTF-8 string, replacing errors to avoid encoding issues
            exc_tb = (
                traceback.format_exc()
                .encode("utf-8", errors="replace")
                .decode("utf-8")
            )
            logger.warning("Metric Function exception caught: %s", exc_tb)
            feedback_result.update(
                error=exc_tb,
                status=feedback_schema.FeedbackResultStatus.FAILED,
            )
            return feedback_result

    def run_and_log(
        self,
        record: record_schema.Record,
        session: core_session.TruSession,
        app: Union[app_schema.AppDefinition, serial_utils.JSON] = None,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
    ) -> Optional[feedback_schema.FeedbackResult]:
        record_id = record.record_id

        db = session.connector.db

        # Placeholder result to indicate a run.
        feedback_result = feedback_schema.FeedbackResult(
            feedback_definition_id=self.feedback_definition_id,
            feedback_result_id=feedback_result_id,
            record_id=record_id,
            name=self.supplied_name
            if self.supplied_name is not None
            else self.name,
        )

        if feedback_result_id is None:
            feedback_result_id = feedback_result.feedback_result_id

        try:
            db.insert_feedback(
                feedback_result.update(
                    status=feedback_schema.FeedbackResultStatus.RUNNING  # in progress
                )
            )

            feedback_result = self.run(app=app, record=record).update(
                feedback_result_id=feedback_result_id
            )

        except Exception:
            # Convert traceback to a UTF-8 string, replacing errors to avoid encoding issues
            exc_tb = (
                traceback.format_exc()
                .encode("utf-8", errors="replace")
                .decode("utf-8")
            )
            db.insert_feedback(
                feedback_result.update(
                    error=exc_tb,
                    status=feedback_schema.FeedbackResultStatus.FAILED,
                )
            )
            return None

        # Otherwise update based on what Metric.run produced (could be success
        # or failure).
        db.insert_feedback(feedback_result)

        return feedback_result

    @property
    def name(self) -> str:
        """Name of the metric.

        Derived from the name of the function implementing it if no supplied
        name provided.
        """

        if self.supplied_name is not None:
            return self.supplied_name

        if self.imp is not None:
            return self.imp.__name__

        return super().name

    def _extract_selection(
        self,
        source_data: Dict,
        combinations: feedback_schema.FeedbackCombinations = feedback_schema.FeedbackCombinations.PRODUCT,
        **kwargs: Dict[str, Any],
    ) -> Iterable[Dict[str, Any]]:
        """
        Create parameter assignments to self.imp from the given data source or
        optionally additional kwargs.

        Args:
            source_data: The data to select from.

            combinations: How to combine assignments for various variables to
                make an assignment to the whole signature.

            **kwargs: Additional keyword arguments to use instead of looking
                them up from source data. Any parameters specified here will be
                used as the assignment value and the selector for that parameter
                will be ignored.

        """

        arg_vals = {}

        for k, q in self.selectors.items():
            try:
                if k in kwargs:
                    arg_vals[k] = [kwargs[k]]
                else:
                    logger.debug(
                        f"Calling q.get with source_data: {source_data}"
                    )
                    result = q.get(source_data)
                    logger.debug(
                        f"Result of q.get(source_data) for key '{k}': {result}"
                    )
                    arg_vals[k] = list(result)
            except Exception as e:
                raise InvalidSelector(
                    selector=q, source_data=source_data
                ) from e

        # For anything specified in kwargs that did not have a selector, set the
        # assignment here as the above loop will have missed it.
        for k, v in kwargs.items():
            if k not in self.selectors:
                arg_vals[k] = [v]

        keys = arg_vals.keys()
        vals = arg_vals.values()

        if combinations == feedback_schema.FeedbackCombinations.PRODUCT:
            assignments = itertools.product(*vals)
        elif combinations == feedback_schema.FeedbackCombinations.ZIP:
            assignments = zip(*vals)
        else:
            raise ValueError(
                f"Unknown combination mode {combinations}. "
                "Expected `product` or `zip`."
            )

        for assignment in assignments:
            yield {k: v for k, v in zip(keys, assignment)}

    def _construct_source_data(
        self,
        app: Optional[
            Union[app_schema.AppDefinition, serial_utils.JSON]
        ] = None,
        record: Optional[record_schema.Record] = None,
        source_data: Optional[Dict] = None,
        **kwargs: dict,
    ) -> Dict:
        """Combine sources of data to be selected over for metric function inputs.

        Args:
            app: The app that produced the record.

            record: The record to evaluate the metric on.

            source_data: Additional data to select from when extracting metric
                function arguments.

            **kwargs: Any additional keyword arguments are merged into
                source_data.

        Returns:
            A dictionary with the combined data.
        """

        if source_data is None:
            source_data = {}
        else:
            source_data = dict(source_data)  # copy

        source_data.update(kwargs)

        if app is not None:
            source_data["__app__"] = app

        if record is not None:
            source_data["__record__"] = record.layout_calls_as_app()

        return source_data

    def extract_selection(
        self,
        app: Optional[
            Union[app_schema.AppDefinition, serial_utils.JSON]
        ] = None,
        record: Optional[record_schema.Record] = None,
        source_data: Optional[Dict] = None,
    ) -> Iterable[Dict[str, Any]]:
        """
        Given the `app` that produced the given `record`, extract from `record`
        the values that will be sent as arguments to the implementation as
        specified by `self.selectors`. Additional data to select from can be
        provided in `source_data`. All args are optional. If a
        [Record][trulens.core.schema.record.Record] is specified, its calls are
        laid out as app (see
        [layout_calls_as_app][trulens.core.schema.record.Record.layout_calls_as_app]).
        """

        return self._extract_selection(
            source_data=self._construct_source_data(
                app=app, record=record, source_data=source_data
            )
        )


Metric.model_rebuild()



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from concurrent import futures
import datetime
import logging
import queue
from threading import Thread
import time
from typing import (
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import pandas as pd
from trulens.core._utils.pycompat import Future  # code style exception
from trulens.core.database import base as core_db
from trulens.core.otel.utils import is_otel_tracing_enabled
from trulens.core.schema import app as app_schema
from trulens.core.schema import event as event_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import types as types_schema
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils

logger = logging.getLogger(__name__)


class DBConnector(ABC, text_utils.WithIdentString):
    """Base class for DB connector implementations."""

    RECORDS_BATCH_TIMEOUT_IN_SEC: int = 10
    """Time to wait before inserting a batch of records into the database."""

    batch_record_queue = queue.Queue()

    batch_thread = None

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.db})"

    # for WithIdentString:
    def _ident_str(self) -> str:
        return self.db._ident_str()

    @property
    @abstractmethod
    def db(self) -> core_db.DB:
        """Get the database instance."""
        ...

    def reset_database(self):
        """Reset the database. Clears all tables.

        See [DB.reset_database][trulens.core.database.base.DB.reset_database].
        """
        self.db.reset_database()

    def migrate_database(self, **kwargs: Any):
        """Migrates the database.

        This should be run whenever there are breaking changes in a database
        created with an older version of _trulens_.

        Args:
            **kwargs: Keyword arguments to pass to
                [migrate_database][trulens.core.database.base.DB.migrate_database]
                of the current database.

        See [DB.migrate_database][trulens.core.database.base.DB.migrate_database].
        """
        self.db.migrate_database(**kwargs)

    def add_record(
        self, record: Optional[record_schema.Record] = None, **kwargs
    ) -> types_schema.RecordID:
        """Add a record to the database.

        Args:
            record: The record to add.

            **kwargs: [Record][trulens.core.schema.record.Record] fields to add to the
                given record or a new record if no `record` provided.

        Returns:
            Unique record identifier [str][] .

        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")

        if record is None:
            record = record_schema.Record(**kwargs)
        else:
            record.update(**kwargs)
        return self.db.insert_record(record=record)

    def add_record_nowait(
        self,
        record: record_schema.Record,
    ) -> None:
        """Add a record to the queue to be inserted in the next batch."""
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        if self.batch_thread is None:
            self.batch_thread = Thread(target=self._batch_loop, daemon=True)
            self.batch_thread.start()
        self.batch_record_queue.put(record)

    def _batch_loop(self):
        apps = {}
        while True:
            time.sleep(self.RECORDS_BATCH_TIMEOUT_IN_SEC)
            records = []
            while True:
                try:
                    record = self.batch_record_queue.get_nowait()
                    records.append(record)
                except queue.Empty:
                    break
            if records:
                try:
                    self.db.batch_insert_record(records)
                except Exception as e:
                    # Re-queue the records that failed to be inserted
                    for record in records:
                        self.batch_record_queue.put(record)
                    logger.error(
                        "Re-queued records due to insertion error {}", e
                    )
                    continue
                feedback_results = []
                for record in records:
                    app_id = record.app_id
                    if app_id not in apps:
                        apps[app_id] = self.get_app(app_id=app_id)
                    app = apps[app_id]

                    feedback_definitions = app.get("feedback_definitions", [])
                    # TODO(Dave): Modify this to add only client side feedback results
                    for feedback_definition_id in feedback_definitions:
                        feedback_results.append(
                            feedback_schema.FeedbackResult(
                                feedback_definition_id=feedback_definition_id,
                                record_id=record.record_id,
                                name="feedback_name",  # this will be updated later by deferred evaluator
                            )
                        )
                try:
                    self.db.batch_insert_feedback(feedback_results)
                except Exception as e:
                    logger.error("Failed to insert feedback results {}", e)

    def add_app(self, app: app_schema.AppDefinition) -> types_schema.AppID:
        """
        Add an app to the database and return its unique id.

        Args:
            app: The app to add to the database.

        Returns:
            A unique app identifier [str][].

        """

        return self.db.insert_app(app=app)

    def delete_app(self, app_id: types_schema.AppID) -> None:
        """
        Deletes an app from the database based on its app_id.

        Args:
            app_id (schema.AppID): The unique identifier of the app to be deleted.
        """
        self.db.delete_app(app_id=app_id)
        logger.info(f"App with ID {app_id} has been successfully deleted.")

    def add_feedback_definition(
        self, feedback_definition: feedback_schema.FeedbackDefinition
    ) -> types_schema.FeedbackDefinitionID:
        """
        Add a feedback definition to the database and return its unique id.

        Args:
            feedback_definition: The feedback definition to add to the database.

        Returns:
            A unique feedback definition identifier [str][].
        """

        return self.db.insert_feedback_definition(
            feedback_definition=feedback_definition
        )

    def add_feedback(
        self,
        feedback_result_or_future: Optional[
            Union[
                feedback_schema.FeedbackResult,
                Future[feedback_schema.FeedbackResult],
            ]
        ] = None,
        **kwargs: Any,
    ) -> types_schema.FeedbackResultID:
        """Add a single feedback result or future to the database and return its unique id.

        Args:
            feedback_result_or_future: If a [Future][concurrent.futures.Future]
                is given, call will wait for the result before adding it to the
                database. If `kwargs` are given and a
                [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] is also
                given, the `kwargs` will be used to update the
                [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] otherwise a
                new one will be created with `kwargs` as arguments to its
                constructor.

            **kwargs: Fields to add to the given feedback result or to create a
                new [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] with.

        Returns:
            A unique result identifier [str][].

        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        if feedback_result_or_future is None:
            if "result" in kwargs and "status" not in kwargs:
                # If result already present, set status to done.
                kwargs["status"] = feedback_schema.FeedbackResultStatus.DONE

            feedback_result = feedback_schema.FeedbackResult(**kwargs)

        elif isinstance(feedback_result_or_future, Future):
            futures.wait([feedback_result_or_future])
            feedback_result: feedback_schema.FeedbackResult = (
                feedback_result_or_future.result()
            )
            feedback_result.update(**kwargs)

        elif isinstance(
            feedback_result_or_future, feedback_schema.FeedbackResult
        ):
            feedback_result = feedback_result_or_future
            feedback_result.update(**kwargs)
        else:
            raise ValueError(
                f"Unknown type {type(feedback_result_or_future)} in feedback_results."
            )

        if feedback_result.feedback_definition_id is None:
            feedback_result.feedback_definition_id = "anonymous"  # or "human" feedback that does not come with a feedback definition

        return self.db.insert_feedback(feedback_result=feedback_result)

    def add_feedbacks(
        self,
        feedback_results: Iterable[
            Union[
                feedback_schema.FeedbackResult,
                Future[feedback_schema.FeedbackResult],
            ]
        ],
    ) -> List[types_schema.FeedbackResultID]:
        """Add multiple feedback results to the database and return their unique ids.
        # TODO: This is slow and should be batched or otherwise optimized in the future.

        Args:
            feedback_results: An iterable with each iteration being a [FeedbackResult][trulens.core.schema.feedback.FeedbackResult] or
                [Future][concurrent.futures.Future] of the same. Each given future will be waited.

        Returns:
            List of unique result identifiers [str][] in the same order as input
                `feedback_results`.
        """
        if is_otel_tracing_enabled():
            raise RuntimeError("Not supported with OTel tracing enabled!")
        return [
            self.add_feedback(
                feedback_result_or_future=feedback_result_or_future
            )
            for feedback_result_or_future in feedback_results
        ]

    def get_app(
        self, app_id: types_schema.AppID
    ) -> Optional[serial_utils.JSONized[app_schema.AppDefinition]]:
        """Look up an app from the database.

        This method produces the JSON-ized version of the app. It can be deserialized back into an [AppDefinition][trulens.core.schema.app.AppDefinition] with [model_validate][pydantic.BaseModel.model_validate]:

        Example:
            ```python
            from trulens.core.schema import app
            app_json = session.get_app(app_id="Custom Application v1")
            app = app.AppDefinition.model_validate(app_json)
            ```

        Warning:
            Do not rely on deserializing into [App][trulens.core.app.App] as
            its implementations feature attributes not meant to be deserialized.

        Args:
            app_id: The unique identifier [str][] of the app to look up.

        Returns:
            JSON-ized version of the app.
        """

        return self.db.get_app(app_id)

    def get_apps(self) -> List[serial_utils.JSONized[app_schema.AppDefinition]]:
        """Look up all apps from the database.

        Returns:
            A list of JSON-ized version of all apps in the database.

        Warning:
            Same Deserialization caveats as [get_app][trulens.core.session.TruSession.get_app].
        """

        return list(self.db.get_apps())

    def get_records_and_feedback(
        self,
        app_ids: Optional[List[types_schema.AppID]] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        app_versions: Optional[List[types_schema.AppVersion]] = None,
        run_name: Optional[types_schema.RunName] = None,
        record_ids: Optional[List[types_schema.RecordID]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Tuple[pd.DataFrame, List[str]]:
        """Get records, their feedback results, and feedback names.

        Args:
            app_ids:
                A list of app ids to filter records by. If empty or not given,
                all apps' records will be returned.
            app_name:
                A name of the app to filter records by. If given, only records
                for this app will be returned.
            app_version:
                A version of the app to filter records by. If given, only
                records for this app version will be returned.
            app_versions:
                A list of app versions to filter records by. If given, only
                records for these app versions will be returned.
            run_name:
                A run name to filter records by. If given, only records for
                this run will be returned.
            record_ids: An optional list of record ids to filter records by.
            offset: Record row offset.
            limit: Limit on the number of records to return.

        Returns:
            Tuple of:
            - DataFrame of records with their feedback results.
            - List of feedback names that are columns in the DataFrame.
        """
        df, feedback_columns = self.db.get_records_and_feedback(
            app_ids=app_ids,
            app_name=app_name,
            app_version=app_version,
            app_versions=app_versions,
            run_name=run_name,
            record_ids=record_ids,
            offset=offset,
            limit=limit,
        )

        df["app_name"] = df["app_json"].apply(lambda x: x.get("app_name"))
        df["app_version"] = df["app_json"].apply(lambda x: x.get("app_version"))

        return df, list(feedback_columns)

    def get_leaderboard(
        self,
        app_ids: Optional[List[types_schema.AppID]] = None,
        group_by_metadata_key: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> pd.DataFrame:
        """Get a leaderboard for the given apps.

        Args:
            app_ids: A list of app ids to filter records by. If empty or not given, all
                apps will be included in leaderboard.

            group_by_metadata_key: A key included in record metadata that you want to group results by.

            limit: Limit on the number of records to aggregate to produce the leaderboard.

            offset: Record row offset to select which records to use to aggregate the leaderboard.

        Returns:
            DataFrame of apps with their feedback results aggregated.
            If group_by_metadata_key is provided, the DataFrame will be grouped by the specified key.
        """

        if app_ids is None:
            app_ids = []

        df, feedback_cols = self.get_records_and_feedback(
            app_ids, limit=limit, offset=offset
        )
        feedback_cols = sorted(feedback_cols)

        df["app_name"] = df["app_json"].apply(lambda x: x.get("app_name"))
        df["app_version"] = df["app_json"].apply(lambda x: x.get("app_version"))

        # TODO: refactor implementation for total_cost map in OTEL implementation of _get_records_and_feedback (see comment: https://github.com/truera/trulens/pull/1939#discussion_r2054802093)
        col_agg_list = feedback_cols + ["latency", "total_cost"]

        if group_by_metadata_key is not None:
            df["meta"] = [df["record_json"][i]["meta"] for i in range(len(df))]

            df[str(group_by_metadata_key)] = [
                item.get(group_by_metadata_key, None)
                if isinstance(item, dict)
                else None
                for item in df["meta"]
            ]
            return (
                df.groupby([
                    "app_name",
                    "app_version",
                    str(group_by_metadata_key),
                ])[col_agg_list]
                .mean()
                .sort_values(by=feedback_cols, ascending=False)
            )
        else:
            return (
                df.groupby(["app_name", "app_version"])[col_agg_list]
                .mean()
                .sort_values(by=feedback_cols, ascending=False)
            )

    def add_event(self, event: event_schema.Event):
        """
        Add an event to the database.

        Args:
            event: The event to add to the database.
        """
        return self.db.insert_event(event=event)

    def add_events(self, events: List[event_schema.Event]):
        """
        Add multiple events to the database.

        Args:
            events: A list of events to add to the database.
        """
        return self.db.insert_events(events=events)

    def get_events(
        self,
        app_name: Optional[str] = None,
        app_version: Optional[str] = None,
        record_ids: Optional[List[str]] = None,
        start_time: Optional[datetime.datetime] = None,
    ) -> pd.DataFrame:
        """
        Get events from the database.

        Args:
            app_name: The app name to filter events by.
            app_version: The app version to filter events by.
            record_ids: The record ids to filter events by.
            start_time: The minimum time to consider events from.

        Returns:
            A pandas DataFrame of all relevant events.
        """
        return self.db.get_events(
            app_name=app_name,
            app_version=app_version,
            record_ids=record_ids,
            start_time=start_time,
        )



import os


def _is_env_var_disabled(var_name: str) -> bool:
    """Check if an environment variable is explicitly set to disable a feature.

    Returns True if the environment variable is explicitly set to "0" or "false"
    (case-insensitive), indicating the feature should be disabled.
    Returns False otherwise (feature enabled by default).
    """
    return os.getenv(var_name, "").lower() in ["0", "false"]


def is_otel_tracing_enabled() -> bool:
    """Check if OpenTelemetry tracing is enabled.

    Returns True by default unless TRULENS_OTEL_TRACING is explicitly set to "0" or "false".
    """
    return not _is_env_var_disabled("TRULENS_OTEL_TRACING")


def is_otel_backwards_compatibility_enabled() -> bool:
    """Check if OpenTelemetry backwards compatibility is enabled.

    Returns True by default unless TRULENS_OTEL_BACKWARDS_COMPATIBILITY is explicitly set to "0" or "false".
    """
    return not _is_env_var_disabled("TRULENS_OTEL_BACKWARDS_COMPATIBILITY")



from dataclasses import dataclass
from typing import Any, Optional


@dataclass
class FeedbackFunctionInput:
    value: Optional[Any] = None
    span_id: Optional[str] = None
    span_attribute: Optional[str] = None
    collect_list: bool = True



"""Utilities related to guessing inputs and outputs of functions."""

import dataclasses
from inspect import BoundArguments
from inspect import Signature
import logging
from typing import Any, Callable, Dict, Sequence

import pydantic
from trulens.core.utils import python as python_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)


def _extract_content(value, content_keys=["content"]):
    """
    Extracts the 'content' from various data types commonly used by libraries
    like OpenAI, Canopy, LiteLLM, etc. This method navigates nested data
    structures (pydantic models, dictionaries, lists) to retrieve the
    'content' field. If 'content' is not directly available, it attempts to
    extract from known structures like 'choices' in a ChatResponse. This
    standardizes extracting relevant text or data from complex API responses
    or internal data representations.

    Args:
        value: The input data to extract content from. Can be a pydantic
                model, dictionary, list, or basic data type.

    Returns:
        The extracted content, which may be a single value, a list of values,
        or a nested structure with content extracted from all levels.
    """
    if isinstance(value, pydantic.BaseModel):
        content = getattr(value, "content", None)
        if content is not None:
            return content

        # If 'content' is not found, check for 'choices' attribute which indicates a ChatResponse
        choices = getattr(value, "choices", None)
        if choices is not None:
            # Extract 'content' from the 'message' attribute of each _Choice in 'choices'
            return [_extract_content(choice.message) for choice in choices]

        # Recursively extract content from nested pydantic models
        try:
            return {
                k: _extract_content(v)
                if isinstance(v, (pydantic.BaseModel, dict, list))
                else v
                for k, v in value.model_dump().items()
            }
        except Exception as e:
            logger.warning(
                "Failed to extract content from pydantic model: %s", e
            )
            # Unsure what is best to do here. Lets just return the exception
            # so that it might reach the user if nothing else gets picked up
            # as main input/output.
            return str(e)

    elif dataclasses.is_dataclass(value) and not isinstance(value, type):
        try:
            value_dict = dataclasses.asdict(value)
            # Check for content keys in the dataclass fields
            for key in content_keys:
                content = value_dict.get(key)
                if content is not None:
                    return content

            # If no specific content key found, recursively extract content from the dictionary
            return {
                k: _extract_content(v) if isinstance(v, (dict, list)) else v
                for k, v in value_dict.items()
            }
        except Exception as e:
            logger.warning("Failed to extract content from dataclass: %s", e)
            # Fall back to string representation
            return str(value)

    elif isinstance(value, dict):
        # Check for 'content' key in the dictionary
        for key in content_keys:
            content = value.get(key)
            if content is not None:
                return content

        # Recursively extract content from nested dictionaries
        return {
            k: _extract_content(v) if isinstance(v, (dict, list)) else v
            for k, v in value.items()
        }

    elif isinstance(value, list):
        # Handle lists by extracting content from each item
        return [_extract_content(item) for item in value]

    else:
        return value


def main_input(func: Callable, sig: Signature, bindings: BoundArguments) -> str:
    """Determine (guess) the main input string for a main app call.

    Args:
        func: The main function we are targeting in this determination.

        sig: The signature of the above.

        bindings: The arguments to be passed to the function.

    Returns:
        The main input string.
    """

    if bindings is None:
        raise RuntimeError(
            f"Cannot determine main input of unbound call to {func}: {sig}."
        )

    # ignore self
    all_args = list(
        v for k, v in bindings.arguments.items() if k not in ["self", "_self"]
    )  # llama_index is using "_self" in some places

    # Helper function to extract meaningful string from a dictionary
    def _extract_from_dict(d):
        # Only extract if we find semantically meaningful keys
        for key in ["content", "input", "query", "text", "message", "name"]:
            if key in d and isinstance(d[key], str):
                return d[key]

        # If no meaningful keys found, don't extract - let it fall through to error handling
        return None

    # Try to find the most meaningful argument by examining each one
    for arg in all_args:
        if arg is None:  # Skip None arguments
            continue

        # Try to extract content from this argument
        extracted = _extract_content(arg, content_keys=["content", "input"])

        # If we got a JSON base type, use it
        if isinstance(extracted, serial_utils.JSON_BASES):
            return str(extracted)

        # If we got a dictionary, try to extract meaningful content
        if isinstance(extracted, Dict):
            result = _extract_from_dict(extracted)
            if result is not None:
                return result

    # Fallback: if have only containers of length 1, find the innermost non-container
    focus = all_args
    while (
        not isinstance(focus, serial_utils.JSON_BASES)
        and isinstance(focus, Sequence)
        and len(focus) == 1
    ):
        focus = focus[0]
        focus = _extract_content(focus, content_keys=["content", "input"])

        if isinstance(focus, serial_utils.JSON_BASES):
            return str(focus)
        if isinstance(focus, Dict):
            result = _extract_from_dict(focus)
            if result is not None:
                return result
        if not isinstance(focus, (Sequence, Dict)):
            logger.warning("Focus %s is not a sequence or dict.", focus)
            break

    if isinstance(focus, serial_utils.JSON_BASES):
        return str(focus)
    if isinstance(focus, Dict):
        result = _extract_from_dict(focus)
        if result is not None:
            return result

    # Otherwise we are not sure.
    logger.warning(
        "Unsure what the main input string is for the call to %s with args %s.",
        python_utils.callable_name(func),
        bindings,
    )

    # After warning, just take the first item in each container until a
    # non-container is reached.
    focus = all_args
    while (
        not isinstance(focus, serial_utils.JSON_BASES)
        and isinstance(focus, Sequence)
        and len(focus) >= 1
    ):
        focus = focus[0]
        focus = _extract_content(focus)

        if not isinstance(focus, (Sequence, Dict)):
            logger.warning("Focus %s is not a sequence or dict.", focus)
            break

    if isinstance(focus, serial_utils.JSON_BASES):
        return str(focus)

    # If we have a dictionary, try to extract meaningful content
    if isinstance(focus, Dict):
        result = _extract_from_dict(focus)
        if result is not None:
            return result

    logger.warning(
        "Could not determine main input/output of %s.", str(all_args)
    )

    return "TruLens: Could not determine main input from " + str(all_args)


def main_output(func: Callable, ret: Any) -> str:
    """Determine (guess) the "main output" string for a given main app call.

    This is for functions whose output is not a string.

    Args:
        func: The main function whose main output we are guessing.

        ret: The return value of the function.
    """

    if isinstance(ret, serial_utils.JSON_BASES):
        return str(ret)

    if isinstance(ret, Sequence) and all(isinstance(x, str) for x in ret):
        # Chunked/streamed outputs.
        return "".join(ret)

    # Use _extract_content to get the content out of the return value
    content = _extract_content(ret, content_keys=["content", "output"])

    if isinstance(content, str):
        return content

    if isinstance(content, float):
        return str(content)

    if isinstance(content, Dict):
        return str(next(iter(content.values()), ""))

    error_message = (
        f"Could not determine main output of {func.__name__}"
        f" from {python_utils.class_name(type(content))} value {content}."
    )

    if isinstance(content, Sequence):
        if len(content) > 0:
            return str(content[0])
        else:
            return error_message

    else:
        logger.warning(error_message)
        return (
            str(content)
            if content is not None
            else (f"TruLens: {error_message}")
        )
